Package: dqcheckrGUI
Type: Package
Title: Point-and-Click GUI Client for 'dqcheckr'
Version: 0.1.0
Date: 2026-06-18
Authors@R: person("Mick", "Mioduszewski",
                  email = "mick@mioduszewski.net",
                  role  = c("aut", "cre"))
Description: 
    A graphical user interface for the 'dqcheckr' package. Provides a
    point-and-click 'shiny' application for configuring dataset quality checks,
    running them against recurring file deliveries, and browsing historical
    check results — without writing any R code.
License: MIT + file LICENSE
URL: https://github.com/mickmioduszewski/dqcheckrGUI
BugReports: https://github.com/mickmioduszewski/dqcheckrGUI/issues
Encoding: UTF-8
Language: en-GB
Depends: R (>= 4.2)
Imports: dqcheckr (>= 0.2.2), shiny, bslib, shinyvalidate, shinyFiles,
        shinyAce, reactable, DT, callr, yaml, readr, DBI, RSQLite,
        stats, tools
Suggests: testthat (>= 3.1.0), knitr, rmarkdown, shinytest2, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-18 20:36:58 UTC; mick
Author: Mick Mioduszewski [aut, cre]
Maintainer: Mick Mioduszewski <mick@mioduszewski.net>
Repository: CRAN
Date/Publication: 2026-06-24 08:20:13 UTC
Built: R 4.5.2; ; 2026-06-24 11:05:25 UTC; unix

# Generated by roxygen2: do not edit by hand

export(run_app)
importFrom(DBI,dbConnect)
importFrom(DT,datatable)
importFrom(RSQLite,SQLite)
importFrom(bslib,bs_theme)
importFrom(callr,r_bg)
importFrom(dqcheckr,run_dq_check)
importFrom(reactable,reactable)
importFrom(readr,cols)
importFrom(shinyAce,aceEditor)
importFrom(shinyFiles,shinyDirButton)
importFrom(shinyvalidate,InputValidator)
importFrom(stats,median)
importFrom(tools,file_path_sans_ext)
importFrom(yaml,read_yaml)

# dqcheckrGUI

A point-and-click Shiny interface for
[dqcheckr](https://github.com/mickmioduszewski/dqcheckr) — automated data
quality checks for recurring dataset deliveries.

Use this if you want to configure and run quality checks without writing any R
code. If you prefer scripting, use `dqcheckr` directly.

## What it does

- Walks you through setting up quality checks for a new dataset (wizard)
- Runs checks against incoming file deliveries and opens the HTML report
- Browses historical check results and drift comparisons

## Installation

```r
# Install dqcheckr first (required)
install.packages("dqcheckr")

# Install dqcheckrGUI from GitHub
devtools::install_github("mickmioduszewski/dqcheckrGUI")
```

## Launch

```r
dqcheckrGUI::run_app()
```

Or, if you have downloaded the source, double-click `launch.command` (Mac),
`launch.sh` (Linux), or `launch.bat` (Windows).

## Example configuration files

Example YAML configuration files using the Star Wars dataset are included with
the package. Copy them to a local folder to use as a starting point:

```r
file.copy(
  system.file("extdata/example_config", package = "dqcheckrGUI"),
  "my_config",
  recursive = TRUE
)
```

## Learn more

See the [dqcheckr documentation](https://mickmioduszewski.github.io/dqcheckr/)
for a full description of all configuration options and quality checks.

# Main application UI (spec §6)

ui <- function() {
  bslib::page_sidebar(
    title = tags$span(style="font-weight:700;font-size:16px;", "dqcheckr"),
    theme = bslib::bs_theme(version=5, bootswatch="flatly", base_font_size="14px"),

    sidebar = bslib::sidebar(
      width = 240,
      style = "padding:0;",

      # Datasets section
      tags$div(
        style="padding:8px 12px 4px;font-size:11px;font-weight:700;text-transform:uppercase;color:#888;letter-spacing:.08em;",
        "Datasets"
      ),
      uiOutput("dataset_sidebar_list"),
      tags$div(
        style="padding:4px 12px 8px;",
        actionButton("btn_new_dataset", "+ New dataset",
                     class="btn btn-sm btn-outline-primary w-100")
      ),
      tags$hr(style="margin:4px 0;"),

      # Other nav buttons
      actionButton("nav_run",     "▶  Run",          class="btn btn-light text-start w-100 border-0 rounded-0 py-2", style="text-align:left;padding-left:16px;"),
      actionButton("nav_history", "⏱  History",      class="btn btn-light text-start w-100 border-0 rounded-0 py-2", style="text-align:left;padding-left:16px;"),
      actionButton("nav_global",  "⚙  Global Config", class="btn btn-light text-start w-100 border-0 rounded-0 py-2", style="text-align:left;padding-left:16px;")
    ),

    # Main panel — content swapped by server
    uiOutput("main_panel"),

    # FWF ruler JS + scroll sync
    tags$head(
      tags$script(src="interact.min.js"),
      tags$script(src="fwf_ruler.js"),
      tags$link(rel="stylesheet", href="app.css")
    ),

    # Global JS handlers — always present in the page
    tags$script('
      Shiny.addCustomMessageHandler("scroll_log", function(msg) {
        var el = document.getElementById("run_log");
        if (el) el.scrollTop = el.scrollHeight;
      });

      // Dataset panel drift-check onchange handler
      window.__dqDC = function(ds) {
        var checked = document.querySelectorAll(".drift-check:checked");
        var btn = document.getElementById("compare_drift_" + ds);
        if (btn) btn.disabled = (checked.length !== 2);
      };

      // Called via onchange= on each history checkbox (inline handler fires before DT)
      window.__dqHC = function(el) {
        var checked = document.querySelectorAll(".hist-check:checked");
        var btn = document.getElementById("history_compare");
        if (!btn) return;
        if (checked.length === 2) {
          var ds1 = checked[0].getAttribute("data-ds");
          var ds2 = checked[1].getAttribute("data-ds");
          if (ds1 === ds2) {
            btn.disabled = false;
            Shiny.setInputValue("hist_selected_ids",
              [checked[0].getAttribute("data-id"), checked[1].getAttribute("data-id")],
              {priority: "event"});
          } else {
            btn.disabled = true;
            Shiny.setInputValue("hist_selected_ids", null, {priority: "event"});
          }
        } else {
          btn.disabled = true;
        }
      };
    ')
  )
}

# dqcheckr client — main entry point (spec §3, §5)

library(shiny)
library(bslib)
library(shinyvalidate)
library(shinyFiles)
library(shinyAce)
library(reactable)
library(DT)
library(callr)
library(yaml)
library(readr)
library(DBI)
library(dqcheckr)

# Source all R/ files into this script's own environment rather than the
# global environment (source()'s `local=FALSE` default writes to .GlobalEnv,
# which would dump every ui_*/server_*/helper function from R/ into the
# user's workspace — a problem for anything sharing the R session, e.g.
# running multiple apps, or tests that load this app via shinytest2/testthat).
# `local=TRUE` sources into the calling frame: the same top-level environment
# in which `ui()` and the `server` closure below are evaluated, so both still
# resolve these definitions via ordinary lexical scoping.
for (f in list.files(file.path(getwd(), "R"), pattern="\\.R$", full.names=TRUE)) {
  source(f, local=TRUE)
}

# Config directory: env var override or working directory
get_config_dir <- function() {
  Sys.getenv("DQCHECKR_CONFIG_DIR", unset=file.path(getwd(), "config"))
}

# Register the reports directory as a static resource path at startup.
# Must be called before shinyApp() so it is available for the first request.
local({
  gcfg <- tryCatch(
    read_global_config(global_config_path(get_config_dir())),
    error = function(e) list()
  )
  register_report_resource_path(gcfg$report_output_dir, get_config_dir())
})

# ── UI ────────────────────────────────────────────────────────────────
shinyApp(
  ui = ui(),

  # ── Server ─────────────────────────────────────────────────────────
  server = function(input, output, session) {

    # Reactive config dir
    config_dir <- reactive({ get_config_dir() })

    # Reactive global config
    gcfg_rv <- reactiveVal(read_global_config(global_config_path(get_config_dir())))

    # Shared reactive state
    rv <- reactiveValues(
      active_section  = "datasets",
      active_dataset  = NULL,
      wizard_open     = FALSE,
      wizard_request  = NULL,
      dataset_refresh = NULL,
      history_refresh = NULL,
      run_preselect   = NULL,
      drift_request   = NULL
    )

    # ── Main panel routing ──────────────────────────────────────────
    output$main_panel <- renderUI({
      s <- rv$active_section
      if (s == "wizard") {
        tagList(
          div(class="d-flex justify-content-between align-items-center mb-2",
            uiOutput("wizard_breadcrumb"),
            actionButton("wizard_cancel", "✕ Cancel",
                         class="btn btn-outline-secondary btn-sm")
          ),
          uiOutput("wizard_step_content"),
          div(class="d-flex gap-2 mt-4",
            uiOutput("wizard_back_btn"),
            uiOutput("wizard_next_btn")
          )
        )
      } else if (s == "datasets") {
        ds <- rv$active_dataset
        if (!is.null(ds)) {
          cfg_path <- file.path(config_dir(), paste0(ds, ".yml"))
          cfg <- if (safe_file_exists(cfg_path)) tryCatch(read_config(cfg_path), error=function(e) list(known=list(), extra=list())) else list(known=list(), extra=list())
          gcfg <- gcfg_rv()
          db_path <- cfg$known$snapshot_db %||% gcfg$snapshot_db %||% ""
          last_runs <- read_snapshot_history(db_path, ds, n=5)
          ui_dataset_panel(ds, cfg, last_runs, gcfg)
        } else {
          tagList(
            h4("Datasets", style="margin-bottom:16px;"),
            p(class="text-muted",
              "Select a dataset from the sidebar, or click ", tags$strong("+ New dataset"), " to create one.")
          )
        }
      } else if (s == "run") {
        datasets <- list_dataset_configs(config_dir())
        ui_run(datasets, selected=rv$run_preselect)
      } else if (s == "history") {
        ui_history()
      } else if (s == "global") {
        gcfg <- gcfg_rv()
        ui_global_config(gcfg)
      }
    })

    # ── Dataset sidebar list ────────────────────────────────────────
    output$dataset_sidebar_list <- renderUI({
      rv$dataset_refresh  # dependency for refresh after save

      datasets <- list_dataset_configs(config_dir())
      if (length(datasets) == 0)
        return(p(class="text-muted px-3", style="font-size:12px;", "No datasets configured."))

      gcfg <- gcfg_rv()
      db_path <- gcfg$snapshot_db %||% ""

      items <- lapply(datasets, function(ds) {
        last_run <- read_snapshot_history(db_path, ds, n=1)
        status_text <- if (nrow(last_run) > 0) last_run$overall_status[1] else ""
        active <- isTRUE(rv$active_dataset == ds)

        tags$div(
          class=paste("dataset-item", if(active) "active"),
          onclick=sprintf("Shiny.setInputValue('sidebar_dataset_click', '%s', {priority:'event'});", js_string_escape(ds)),
          span(ds, style="flex:1;"),
          if (nchar(status_text) > 0) status_badge(status_text)
        )
      })
      tagList(items)
    })

    # ── Sidebar navigation ──────────────────────────────────────────
    observeEvent(input$sidebar_dataset_click, {
      rv$active_dataset <- input$sidebar_dataset_click
      rv$active_section <- "datasets"
    })
    observeEvent(input$btn_new_dataset, {
      rv$wizard_request <- list(mode="new", dataset=NULL, ts=Sys.time())
    })
    observeEvent(input$nav_run, {
      rv$active_section <- "run"
      rv$run_preselect <- rv$active_dataset
    })
    observeEvent(input$nav_history, {
      rv$active_section <- "history"
      rv$history_refresh <- Sys.time()
    })
    observeEvent(input$nav_global, {
      rv$active_section <- "global"
    })

    # ── Dataset panel actions (edit, run, history, compare) ─────────
    observeEvent(input$ds_action, {
      req(is.list(input$ds_action))
      action <- input$ds_action$action
      ds     <- input$ds_action$ds
      if (action == "edit") {
        rv$wizard_request <- list(mode="edit", dataset=ds, ts=Sys.time())
      } else if (action == "run") {
        rv$run_preselect  <- ds
        rv$active_section <- "run"
      } else if (action == "history") {
        rv$active_section  <- "history"
        rv$history_refresh <- Sys.time()
      } else if (action == "compare") {
        ids <- input$ds_action$ids
        if (length(ids) == 2) {
          rv$drift_request <- list(ds = ds, ids = ids, ts = Sys.time())
        }
      }
    })

    # ── Wizard cancel ───────────────────────────────────────────────
    observeEvent(input$wizard_cancel, {
      showModal(modalDialog(
        title="Cancel wizard?",
        "Unsaved changes will be lost.",
        footer=tagList(
          modalButton("Keep editing"),
          actionButton("wizard_cancel_confirm", "Discard and exit", class="btn btn-danger")
        )
      ))
    })
    observeEvent(input$wizard_cancel_confirm, {
      removeModal()
      rv$wizard_open <- FALSE
      rv$active_section <- "datasets"
    })

    # ── Initialise sub-servers ──────────────────────────────────────
    wiz <- server_wizard(input, output, session, rv, config_dir, gcfg_rv)
    server_step3_csv(input, output, session, wiz)
    server_global(input, output, session, rv, config_dir, gcfg_rv)
    server_run(input, output, session, rv, config_dir, gcfg_rv)
    server_history(input, output, session, rv, config_dir, gcfg_rv)

    # ── First-run: open global config if no dqcheckr.yml ───────────
    observe({
      cfg_path <- global_config_path(config_dir())
      if (!safe_file_exists(cfg_path)) {
        rv$active_section <- "global"
        showNotification(
          "Welcome! Please configure the infrastructure paths below to get started.",
          type="message", duration=6
        )
      }
    })

  }
)

# History panel server logic (spec §18)

server_history <- function(input, output, session, rv, config_dir, gcfg_rv) {

  hist_data  <- reactiveVal(data.frame())
  hist_limit <- reactiveVal(50L)

  # Load / reload history
  load_history <- function() {
    db_path <- resolve_infra_path(
      gcfg_rv()$snapshot_db, config_dir(),
      default = "data/snapshots.sqlite", mustWork = FALSE
    )
    tryCatch({
      df <- read_all_snapshot_history(db_path, n = hist_limit())
      hist_data(df)
    }, error = function(e) hist_data(data.frame()))
  }

  # Refresh when section activated or after a run
  observe({
    rv$active_section
    rv$history_refresh
    load_history()
  })

  observeEvent(input$history_load_more, {
    hist_limit(hist_limit() + 50L)
    load_history()
  })

  output$history_table <- DT::renderDataTable({
    df <- hist_data()
    if (nrow(df) == 0) {
      return(DT::datatable(
        data.frame(Message = "No run history found. Run a quality check first."),
        options = list(dom = "t"), rownames = FALSE
      ))
    }

    # Build direct /dq_reports/<file> links — avoids Shiny round-trip and popup blocking
    filename <- make_report_filename(df$dataset_name, df$run_timestamp)

    display_df <- data.frame(
      ` ` = sprintf(
        '<input type="checkbox" class="hist-check" data-id="%d" data-ds="%s" onchange="window.__dqHC(this)"/>',
        df$id, df$dataset_name),
      Dataset = df$dataset_name,
      Date    = utc_to_local_display(df$run_timestamp),
      File    = df$file_name,
      Status  = vapply(df$overall_status, status_badge_html, character(1)),
      Fails   = df$check_fail_count,
      Rows    = df$row_count,
      Report  = sprintf(
        '<a href="javascript:void(0)" onclick="window.open(\'/dq_reports/%s\',\'_blank\')">Open</a>',
        filename),
      stringsAsFactors = FALSE, check.names = FALSE
    )
    DT::datatable(display_df,
      escape    = FALSE, rownames = FALSE, selection = "none",
      filter    = "top",
      options   = list(pageLength = 50, scrollX = TRUE, dom = "ftip"),
      class     = "compact stripe hover"
    )
  })

  # Shared drift-launch helper
  launch_drift <- function(ds_name, id1, id2) {
    cd         <- config_dir()
    db_path    <- resolve_infra_path(gcfg_rv()$snapshot_db, cd,
                                     default = "data/snapshots.sqlite", mustWork = FALSE)
    report_dir <- resolve_infra_path(gcfg_rv()$report_output_dir, cd,
                                     default = "reports/", mustWork = FALSE)
    prev_id    <- min(as.integer(c(id1, id2)))
    curr_id    <- max(as.integer(c(id1, id2)))
    showNotification("Starting drift comparison...", type = "message", duration = 4)
    drift_log           <- tempfile("dqcheckr_drift_", fileext = ".log")
    rv_drift$proc       <- callr::r_bg(
      func = function(dn, p, c, db, cd) {
        dqcheckr::compare_snapshots(dn,
          snapshot_id_prev = p, snapshot_id_curr = c,
          db_path = db, config_dir = cd,
          report = TRUE, open_report = FALSE)
      },
      args    = list(dn = ds_name, p = prev_id, c = curr_id, db = db_path, cd = cd),
      stdout  = drift_log,
      stderr  = "2>&1",
      # Run in the deployment root so dqcheckr resolves any relative paths the
      # same way the CLI does (shiny::runApp has changed this process's getwd()).
      wd      = deployment_root(cd),
      package = TRUE
    )
    rv_drift$log_path   <- drift_log
    rv_drift$report_dir <- report_dir
    rv_drift$ds_name    <- ds_name
  }

  # Compare drift — async, non-blocking
  rv_drift <- reactiveValues(proc = NULL, report_dir = NULL, ds_name = NULL, log_path = NULL)

  observeEvent(input$history_compare, {
    ids <- input$hist_selected_ids
    req(length(ids) == 2)
    df   <- hist_data()
    row1 <- df[df$id == as.integer(ids[1]), ]
    req(nrow(row1) > 0)
    launch_drift(row1$dataset_name[1], ids[1], ids[2])
  })

  # Drift request from dataset panel compare button
  observeEvent(rv$drift_request, {
    req(!is.null(rv$drift_request))
    req(length(rv$drift_request$ids) == 2)
    launch_drift(rv$drift_request$ds, rv$drift_request$ids[1], rv$drift_request$ids[2])
  })

  # Poll for drift completion — 500ms intervals, non-blocking
  observe({
    req(!is.null(rv_drift$proc))
    invalidateLater(500)
    proc <- rv_drift$proc
    if (is.null(proc) || proc$is_alive()) return()

    result <- tryCatch(proc$get_result(), error = function(e) {
      last_line <- tryCatch({
        lp <- rv_drift$log_path
        if (!is.null(lp) && file.exists(lp)) {
          lines <- readLines(lp, warn = FALSE)
          lines <- lines[nchar(trimws(lines)) > 0]
          if (length(lines) > 0) tail(lines, 1) else ""
        } else ""
      }, error = function(e2) "")
      msg <- if (nchar(last_line) > 0) last_line else conditionMessage(e)
      showNotification(paste("Drift comparison failed:", msg),
                       type = "error", duration = 10)
      NULL
    })
    rv_drift$proc <- NULL
    if (is.null(result)) return()

    ds_name     <- rv_drift$ds_name
    # Anchor to the exact "drift_<dataset>_<YYYYMMDD>_<HHMMSS>.html" slug shape
    # (see make_report_filename(), utils.R) rather than a loose ".*" — a bare
    # "^drift_<ds_name>_.*\\.html$" would also match e.g. "RBB_bonds"'s report
    # when ds_name is "RBB", since "bonds_<timestamp>.html" satisfies ".*\\.html$".
    drift_files <- list.files(rv_drift$report_dir,
                              pattern    = sprintf("^drift_%s_[0-9]{8}_[0-9]{6}\\.html$", ds_name),
                              full.names = FALSE)
    if (length(drift_files) > 0) {
      url <- paste0("/dq_reports/", drift_files[length(drift_files)])
      showModal(modalDialog(
        title  = "Drift comparison complete",
        tags$a(href = url, target = "_blank", class = "btn btn-primary",
               "Open drift report ↗"),
        footer = modalButton("Close"),
        easyClose = TRUE
      ))
    } else {
      showNotification(
        "Drift comparison complete. Report not found in reports directory.",
        type = "warning"
      )
    }
  })
}

#' dqcheckrGUI: Point-and-Click GUI Client for 'dqcheckr'
#'
#' A point-and-click 'shiny' front-end for \pkg{dqcheckr} -- configure and run
#' automated data quality checks on recurring dataset deliveries without
#' writing any R code.
#'
#' The entry point is \code{\link{run_app}}.
#'
#' These packages are only called from the 'shiny' app sourced at runtime via
#' \code{system.file("app", package = "dqcheckrGUI")}
#' (\code{inst/app/R/*.R}), so static analysis of \code{R/} cannot see them
#' as used -- without a reference here, \code{R CMD check} reports
#' "Namespaces in Imports field not imported from".
#' @importFrom DBI dbConnect
#' @importFrom DT datatable
#' @importFrom RSQLite SQLite
#' @importFrom bslib bs_theme
#' @importFrom callr r_bg
#' @importFrom dqcheckr run_dq_check
#' @importFrom reactable reactable
#' @importFrom readr cols
#' @importFrom shinyAce aceEditor
#' @importFrom shinyFiles shinyDirButton
#' @importFrom shinyvalidate InputValidator
#' @importFrom stats median
#' @importFrom tools file_path_sans_ext
#' @importFrom yaml read_yaml
#'
#' @keywords internal
#' @noRd
"_PACKAGE"

# Dataset list and dataset panel UI (spec §7)

ui_dataset_panel <- function(dataset_name, config, last_runs, gcfg) {
  format_str  <- toupper(config$known$format %||% "CSV")
  loc_str     <- if (nchar(config$known$folder %||% "") > 0)
                   config$known$folder
                 else config$known$current_file %||% "(not set)"
  # Escaped once for every onclick JS-string interpolation below — see
  # js_string_escape() for why this matters even though is_valid_r_name()
  # constrains names created via the wizard.
  ds_js <- js_string_escape(dataset_name)

  tagList(
    div(class="d-flex align-items-center gap-3 mb-3",
      h4(dataset_name, style="margin:0;"),
      span(style="font-size:12px;background:#e9ecef;padding:2px 8px;border-radius:4px;", format_str)
    ),
    p(class="text-muted", style="font-size:13px;margin-bottom:4px;",
      tags$strong("Location: "), loc_str),

    div(class="d-flex gap-2 mt-3 mb-4",
      tags$button("Edit config",
        class="btn btn-outline-secondary btn-sm",
        onclick=sprintf("Shiny.setInputValue('ds_action', {action:'edit', ds:'%s', ts:Date.now()}, {priority:'event'});", ds_js)),
      tags$button("▶ Run check",
        class="btn btn-primary btn-sm",
        onclick=sprintf("Shiny.setInputValue('ds_action', {action:'run', ds:'%s', ts:Date.now()}, {priority:'event'});", ds_js))
    ),

    h6("Recent runs", class="text-muted"),
    if (nrow(last_runs) == 0) {
      p(class="text-muted fst-italic", "No runs recorded yet.")
    } else {
      local({
        filename <- make_report_filename(dataset_name, last_runs$run_timestamp)
        tagList(
          DT::datatable(
            data.frame(
              ` ` = sprintf(
                '<input type="checkbox" class="drift-check" data-id="%d" data-ds="%s" onchange="window.__dqDC(\'%s\')"/>',
                last_runs$id, dataset_name, ds_js),
              Date   = utc_to_local_display(last_runs$run_timestamp),
              File   = last_runs$file_name,
              Status = vapply(last_runs$overall_status, status_badge_html, character(1)),
              Fails  = last_runs$check_fail_count,
              Report = sprintf(
                '<a href="javascript:void(0)" onclick="window.open(\'/dq_reports/%s\',\'_blank\')">Open</a>',
                filename),
              stringsAsFactors = FALSE, check.names = FALSE
            ),
            escape = FALSE, rownames = FALSE, selection = "none",
            options = list(dom="t", ordering=FALSE, pageLength=5, scrollX=TRUE),
            class = "compact stripe hover"
          ),
          div(class="d-flex gap-2 mt-2 align-items-center",
            tags$button("Compare drift ▶",
              id    = paste0("compare_drift_", dataset_name),
              class = "btn btn-outline-info btn-sm",
              disabled = "",
              onclick = sprintf(
                "var ids=Array.from(document.querySelectorAll('.drift-check:checked')).map(function(c){return c.getAttribute('data-id');});Shiny.setInputValue('ds_action',{action:'compare',ds:'%s',ids:ids,ts:Date.now()},{priority:'event'});",
                ds_js)),
            span(class="text-muted fst-italic", style="font-size:12px;",
                 "Select 2 runs above to compare")
          )
        )
      })
    },
    div(class="mt-2",
      tags$a("View all in History →", href="#", class="btn btn-link btn-sm p-0",
        onclick=sprintf("Shiny.setInputValue('ds_action', {action:'history', ds:'%s', ts:Date.now()}, {priority:'event'}); return false;", ds_js))
    )
  )
}

# YAML config round-trip with unknown-key preservation (spec §20)

KNOWN_KEYS <- c(
  "dataset_name","format","encoding","delimiter","quote_char","col_names","csv_skip","folder",
  "current_file","previous_file","fwf_widths","fwf_col_names","fwf_skip",
  "expected_columns","key_columns","column_types","column_rules",
  "rule_overrides","custom_checks_file","snapshot_db","report_output_dir",
  "description"
)

read_config <- function(path) {
  raw <- yaml::read_yaml(path)
  known <- list(
    dataset_name       = raw$dataset_name %||% "",
    description        = raw$description %||% "",
    format             = raw$format %||% "csv",
    encoding           = raw$encoding %||% "UTF-8",
    delimiter          = raw$delimiter %||% ",",
    quote_char         = raw$quote_char %||% '"',
    # Header presence is the conjunction of two orthogonal facts. col_names may
    # be present either because the file is genuinely headerless (csv_skip 0) OR
    # because a usable-but-replaced header is being skipped (csv_skip >= 1). Only
    # the former means "no header". Without this 3-way rule, a renamed-header
    # config would round-trip to has_header = FALSE and lose csv_skip on re-save.
    has_header         = is.null(raw$col_names) || (raw$csv_skip %||% 0L) >= 1L,
    col_names          = if (!is.null(raw$col_names)) unlist(raw$col_names) else character(0),
    csv_skip           = as.integer(raw$csv_skip %||% 0L),
    folder             = raw$folder %||% "",
    current_file       = raw$current_file %||% "",
    previous_file      = raw$previous_file %||% "",
    fwf_widths         = raw$fwf_widths %||% integer(0),
    fwf_col_names      = raw$fwf_col_names %||% character(0),
    fwf_skip           = raw$fwf_skip %||% 0L,
    expected_columns   = raw$expected_columns %||% character(0),
    key_columns        = raw$key_columns %||% character(0),
    col_types_override = raw$column_types %||% list(),
    column_rules       = raw$column_rules %||% list(),
    rule_overrides     = raw$rule_overrides %||% list(),
    custom_checks_file = raw$custom_checks_file %||% "",
    snapshot_db        = raw$snapshot_db %||% "",
    report_output_dir  = raw$report_output_dir %||% ""
  )
  # file_mode derived
  known$file_mode <- if (nchar(known$folder) > 0) "folder" else "explicit"
  # Preserve unknown keys
  extra <- raw[setdiff(names(raw), KNOWN_KEYS)]
  list(known = known, extra = extra)
}

read_global_config <- function(path) {
  if (!file.exists(path)) {
    return(list(
      snapshot_db       = "data/snapshots.sqlite",
      report_output_dir = "reports/",
      default_rules = list(
        type_inference_threshold      = 0.90,
        max_missing_rate              = 0.05,
        max_non_numeric_rate          = 0.01,
        min_row_count                 = 0,
        max_row_count_change_pct      = 0.10,
        max_numeric_mean_shift_pct    = 0.20,
        max_missing_rate_change_pp    = 2.0,
        max_non_numeric_rate_change_pp= 1.0,
        flag_new_columns              = TRUE,
        flag_dropped_columns          = TRUE,
        flag_type_changes             = TRUE,
        flag_column_order_change      = TRUE
      )
    ))
  }
  yaml::read_yaml(path)
}

write_global_config <- function(values, path) {
  yaml::write_yaml(values, path)
}

build_config_list <- function(wiz) {
  cfg <- list()
  cfg$dataset_name <- wiz$dataset_name
  if (nchar(wiz$description %||% "") > 0) cfg$description <- wiz$description

  cfg$format   <- wiz$format
  cfg$encoding <- wiz$encoding

  if (wiz$format == "csv") {
    cfg$delimiter <- wiz$delimiter
    if (!is.null(wiz$quote_char) && wiz$quote_char != '"')
      cfg$quote_char <- wiz$quote_char

    if (isFALSE(wiz$has_header)) {
      # Genuinely headerless: names are user-supplied, no header line to skip.
      if (length(wiz$col_names) > 0)
        cfg$col_names <- as.list(wiz$col_names)
    } else {
      # Header present. Emit col_names + csv_skip = 1 ONLY when the names were
      # changed away from the file's actual header (duplicate/invalid fixes or
      # deliberate renames). A clean header — or one we couldn't probe — writes
      # neither key, keeping the common case byte-identical to legacy output.
      raw <- as.character(wiz$raw_header_names %||% character(0))
      renamed <- length(raw) > 0 &&
                 length(wiz$col_names) == length(raw) &&
                 !identical(as.character(wiz$col_names), raw)
      if (renamed) {
        cfg$col_names <- as.list(wiz$col_names)
        cfg$csv_skip  <- 1L
      }
    }
  }

  if (wiz$file_mode == "folder") {
    cfg$folder <- wiz$folder
  } else {
    cfg$current_file <- wiz$current_file
    if (nchar(wiz$previous_file %||% "") > 0)
      cfg$previous_file <- wiz$previous_file
  }

  if (wiz$format == "fwf") {
    cfg$fwf_widths    <- as.list(wiz$fwf_widths)
    cfg$fwf_col_names <- as.list(wiz$fwf_col_names)
    cfg$fwf_skip      <- as.integer(wiz$fwf_skip)
  }

  if (length(wiz$expected_columns) > 0)
    cfg$expected_columns <- as.list(wiz$expected_columns)

  if (length(wiz$key_columns) > 0)
    cfg$key_columns <- as.list(wiz$key_columns)

  if (length(wiz$col_types_override) > 0)
    cfg$column_types <- wiz$col_types_override

  # Build column_rules — only include columns that have at least one rule
  col_rules <- list()
  for (col in names(wiz$column_rules)) {
    rules <- wiz$column_rules[[col]]
    rules <- rules[!sapply(rules, function(x) is.null(x) || (length(x)==1 && (is.na(x) || x=="")))]
    if (length(rules) > 0) col_rules[[col]] <- rules
  }
  if (length(col_rules) > 0) cfg$column_rules <- col_rules

  # Rule overrides — only non-default values
  if (length(wiz$rule_overrides) > 0) {
    overrides <- wiz$rule_overrides[!sapply(wiz$rule_overrides, is.null)]
    if (length(overrides) > 0) cfg$rule_overrides <- overrides
  }

  if (nchar(wiz$custom_checks_file %||% "") > 0)
    cfg$custom_checks_file <- wiz$custom_checks_file

  # Per-dataset infrastructure path overrides (only if set and non-empty)
  if (nchar(wiz$snapshot_db %||% "") > 0)
    cfg$snapshot_db <- wiz$snapshot_db
  if (nchar(wiz$report_output_dir %||% "") > 0)
    cfg$report_output_dir <- wiz$report_output_dir

  cfg
}

write_config <- function(wiz, extra, path) {
  cfg <- build_config_list(wiz)
  # Merge extra keys (unknown keys from original file)
  for (key in names(extra)) {
    if (!key %in% names(cfg)) cfg[[key]] <- extra[[key]]
  }
  yaml::write_yaml(cfg, path)
}

yaml_preview_text <- function(wiz, extra) {
  cfg <- build_config_list(wiz)
  main_yaml <- yaml::as.yaml(cfg)
  if (length(extra) > 0) {
    extra_yaml <- yaml::as.yaml(extra)
    paste0(main_yaml, "\n# preserved from original file\n", extra_yaml)
  } else {
    main_yaml
  }
}

# Unit tests for R/config_io.R
# These catch YAML round-trip bugs without needing a running Shiny app.

library(yaml)

# ── read_config: field reading ───────────────────────────────────────────────

test_that("read_config reads all standard fields", {
  path <- write_yaml_fixture(list(
    dataset_name     = "bonds",
    description      = "My dataset",
    format           = "csv",
    encoding         = "UTF-8",
    delimiter        = ",",
    current_file     = "/data/file.csv",
    expected_columns = list("id", "name", "value"),
    key_columns      = list("id"),
    column_types     = list(value = "numeric"),
    column_rules     = list(value = list(min_value = 0, max_value = 999)),
    rule_overrides   = list(max_missing_rate = 0.10),
    custom_checks_file = "/checks/my_checks.R"
  ))

  k <- read_config(path)$known

  expect_equal(k$dataset_name,       "bonds")
  expect_equal(k$description,        "My dataset")
  expect_equal(k$format,             "csv")
  expect_equal(k$encoding,           "UTF-8")
  expect_equal(k$delimiter,          ",")
  expect_equal(k$current_file,       "/data/file.csv")
  expect_equal(k$expected_columns,   c("id", "name", "value"))
  expect_equal(k$key_columns,        c("id"))
  expect_equal(k$col_types_override, list(value = "numeric"))
  expect_equal(k$column_rules$value$min_value, 0)
  expect_equal(k$column_rules$value$max_value, 999)
  expect_equal(k$rule_overrides$max_missing_rate, 0.10)
  expect_equal(k$custom_checks_file, "/checks/my_checks.R")
})

test_that("read_config provides sensible defaults for missing fields", {
  path <- write_yaml_fixture(list(dataset_name = "minimal"))
  k <- read_config(path)$known

  expect_equal(k$format,    "csv")
  expect_equal(k$encoding,  "UTF-8")
  expect_equal(k$delimiter, ",")
  expect_equal(k$folder,    "")
  expect_equal(k$current_file, "")
  expect_equal(k$key_columns, character(0))
  expect_equal(k$expected_columns, character(0))
  expect_equal(k$col_types_override, list())
  expect_equal(k$column_rules, list())
  expect_equal(k$rule_overrides, list())
})

# ── read_config: file_mode derivation ────────────────────────────────────────

test_that("read_config derives file_mode=explicit when only current_file set", {
  path <- write_yaml_fixture(list(dataset_name = "ds", current_file = "/data/f.csv"))
  expect_equal(read_config(path)$known$file_mode, "explicit")
})

test_that("read_config derives file_mode=folder when folder is set", {
  path <- write_yaml_fixture(list(dataset_name = "ds", folder = "/data/folder"))
  expect_equal(read_config(path)$known$file_mode, "folder")
})

test_that("read_config derives file_mode=explicit when folder is empty string", {
  path <- write_yaml_fixture(list(dataset_name = "ds", folder = "", current_file = "/f.csv"))
  expect_equal(read_config(path)$known$file_mode, "explicit")
})

# ── read_config: unknown key preservation ────────────────────────────────────

test_that("read_config preserves unknown keys in $extra", {
  path <- write_yaml_fixture(list(
    dataset_name   = "ds",
    org_flag       = TRUE,
    project_id     = "XYZ-42",
    internal_notes = "do not remove"
  ))
  result <- read_config(path)

  expect_true("org_flag"       %in% names(result$extra))
  expect_true("project_id"     %in% names(result$extra))
  expect_true("internal_notes" %in% names(result$extra))
  expect_equal(result$extra$project_id, "XYZ-42")
})

test_that("read_config does not put unknown keys in $known", {
  path <- write_yaml_fixture(list(dataset_name = "ds", my_custom = "value"))
  expect_false("my_custom" %in% names(read_config(path)$known))
})

test_that("read_config returns empty extra when all keys are known", {
  path <- write_yaml_fixture(list(dataset_name = "ds", format = "csv"))
  expect_equal(length(read_config(path)$extra), 0L)
})

# ── build_config_list: file mode branching ───────────────────────────────────
# These tests encode the bug where wiz$file_mode wasn't synced from the radio
# button, causing current_file to be silently dropped from the YAML.

test_that("build_config_list writes current_file (not folder) in explicit mode", {
  wiz <- make_wiz(file_mode = "explicit", current_file = "/data/delivery.csv")
  cfg <- build_config_list(wiz)

  expect_equal(cfg$current_file, "/data/delivery.csv")
  expect_null(cfg$folder)  # folder must NOT appear
})

test_that("build_config_list writes folder (not current_file) in folder mode", {
  wiz <- make_wiz(file_mode = "folder", folder = "/data/monthly/")
  cfg <- build_config_list(wiz)

  expect_equal(cfg$folder, "/data/monthly/")
  expect_null(cfg$current_file)  # current_file must NOT appear
})

test_that("build_config_list includes previous_file when set in explicit mode", {
  wiz <- make_wiz(file_mode = "explicit",
                  current_file = "/data/curr.csv",
                  previous_file = "/data/prev.csv")
  cfg <- build_config_list(wiz)

  expect_equal(cfg$previous_file, "/data/prev.csv")
})

test_that("build_config_list omits previous_file when empty", {
  wiz <- make_wiz(file_mode = "explicit", current_file = "/data/curr.csv", previous_file = "")
  cfg <- build_config_list(wiz)

  expect_null(cfg$previous_file)
})

# ── build_config_list: description ───────────────────────────────────────────

test_that("build_config_list omits description when empty", {
  expect_null(build_config_list(make_wiz(description = ""))$description)
})

test_that("build_config_list includes description when set", {
  expect_equal(build_config_list(make_wiz(description = "My bonds data"))$description,
               "My bonds data")
})

# ── build_config_list: column metadata ───────────────────────────────────────

test_that("build_config_list includes expected_columns and key_columns", {
  wiz <- make_wiz(
    expected_columns = c("id", "name", "amount"),
    key_columns      = c("id")
  )
  cfg <- build_config_list(wiz)

  expect_equal(unlist(cfg$expected_columns), c("id", "name", "amount"))
  expect_equal(unlist(cfg$key_columns),      c("id"))
})

test_that("build_config_list omits expected_columns and key_columns when empty", {
  wiz <- make_wiz(expected_columns = character(0), key_columns = character(0))
  cfg <- build_config_list(wiz)

  expect_null(cfg$expected_columns)
  expect_null(cfg$key_columns)
})

test_that("build_config_list writes column_types overrides", {
  wiz <- make_wiz(col_types_override = list(code = "character", amount = "numeric"))
  cfg <- build_config_list(wiz)

  expect_equal(cfg$column_types$code,   "character")
  expect_equal(cfg$column_types$amount, "numeric")
})

test_that("build_config_list omits column_types when empty", {
  expect_null(build_config_list(make_wiz(col_types_override = list()))$column_types)
})

test_that("build_config_list strips column_rules entries that are all empty", {
  wiz <- make_wiz(column_rules = list(
    id    = list(min_value = NA, max_value = NULL, pattern = ""),
    value = list(min_value = 0, max_value = 1000)
  ))
  cfg <- build_config_list(wiz)

  expect_null(cfg$column_rules$id)
  expect_equal(cfg$column_rules$value$min_value, 0)
})

test_that("build_config_list omits column_rules entirely when all empty", {
  wiz <- make_wiz(column_rules = list(id = list(min_value = NA, pattern = "")))
  expect_null(build_config_list(wiz)$column_rules)
})

# ── write_config: full round-trip ────────────────────────────────────────────

test_that("write_config round-trip preserves all known fields", {
  wiz <- make_wiz(
    dataset_name     = "round_trip",
    description      = "Test dataset",
    file_mode        = "explicit",
    current_file     = "/data/file.csv",
    expected_columns = c("a", "b", "c"),
    key_columns      = c("a"),
    col_types_override = list(b = "numeric"),
    column_rules     = list(b = list(min_value = 0, max_value = 999)),
    rule_overrides   = list(max_missing_rate = 0.10)
  )
  path <- tempfile(fileext = ".yml")
  write_config(wiz, list(), path)

  k <- read_config(path)$known
  expect_equal(k$dataset_name,   "round_trip")
  expect_equal(k$description,    "Test dataset")
  expect_equal(k$current_file,   "/data/file.csv")
  expect_equal(k$file_mode,      "explicit")
  expect_equal(k$expected_columns, c("a", "b", "c"))
  expect_equal(k$key_columns,      c("a"))
  expect_equal(k$col_types_override, list(b = "numeric"))
  expect_equal(k$column_rules$b$min_value, 0)
  expect_equal(k$rule_overrides$max_missing_rate, 0.10)
})

test_that("write_config preserves unknown keys through round-trip", {
  original <- list(
    dataset_name   = "ds",
    current_file   = "/data/f.csv",
    org_flag       = TRUE,
    project_id     = "XYZ"
  )
  path1 <- write_yaml_fixture(original)

  result <- read_config(path1)
  wiz    <- make_wiz(dataset_name = result$known$dataset_name,
                     file_mode    = result$known$file_mode,
                     current_file = result$known$current_file)

  path2 <- tempfile(fileext = ".yml")
  write_config(wiz, result$extra, path2)

  raw <- yaml::read_yaml(path2)
  expect_equal(raw$org_flag,   TRUE)
  expect_equal(raw$project_id, "XYZ")
  expect_equal(raw$current_file, "/data/f.csv")
})

test_that("write_config does not duplicate unknown keys as known fields", {
  original <- list(dataset_name = "ds", current_file = "/f.csv", extra_key = "v")
  path1    <- write_yaml_fixture(original)
  result   <- read_config(path1)

  wiz   <- make_wiz(file_mode = "explicit", current_file = "/f.csv")
  path2 <- tempfile(fileext = ".yml")
  write_config(wiz, result$extra, path2)

  raw <- yaml::read_yaml(path2)
  # extra_key should appear exactly once
  expect_equal(sum(names(raw) == "extra_key"), 1L)
})

# ── csv_skip / has_header inference (duplicate-header support, B-08) ───────────

test_that("read_config: col_names + csv_skip>=1 means the file HAS a header", {
  path <- write_yaml_fixture(list(
    dataset_name = "refunds",
    current_file = "/data/refunds.csv",
    col_names    = list("PayeeName", "PayeeName_2", "Amount"),
    csv_skip     = 1L
  ))
  k <- read_config(path)$known
  expect_true(k$has_header)                       # NOT headerless
  expect_equal(k$csv_skip, 1L)
  expect_equal(k$col_names, c("PayeeName", "PayeeName_2", "Amount"))
})

test_that("read_config: col_names without csv_skip means genuinely headerless", {
  path <- write_yaml_fixture(list(
    dataset_name = "nohdr",
    current_file = "/data/nohdr.csv",
    col_names    = list("a", "b", "c")
  ))
  k <- read_config(path)$known
  expect_false(k$has_header)
  expect_equal(k$csv_skip, 0L)
})

test_that("read_config: no col_names means clean header (defaults)", {
  path <- write_yaml_fixture(list(dataset_name = "clean", current_file = "/f.csv"))
  k <- read_config(path)$known
  expect_true(k$has_header)
  expect_equal(k$csv_skip, 0L)
  expect_equal(k$col_names, character(0))
})

test_that("build_config_list: renamed header writes col_names + csv_skip=1", {
  wiz <- make_wiz(
    file_mode        = "explicit", current_file = "/data/refunds.csv",
    has_header       = TRUE,
    raw_header_names = c("PayeeName", "PayeeName", "Amount"),   # file's real header
    col_names        = c("PayeeName", "PayeeName_2", "Amount")  # user/suggested fix
  )
  cfg <- build_config_list(wiz)
  expect_equal(unlist(cfg$col_names), c("PayeeName", "PayeeName_2", "Amount"))
  expect_equal(cfg$csv_skip, 1L)
})

test_that("build_config_list: clean header writes neither col_names nor csv_skip", {
  wiz <- make_wiz(
    file_mode        = "explicit", current_file = "/data/clean.csv",
    has_header       = TRUE,
    raw_header_names = c("id", "name", "amount"),
    col_names        = c("id", "name", "amount")               # unchanged
  )
  cfg <- build_config_list(wiz)
  expect_null(cfg$col_names)
  expect_null(cfg$csv_skip)
})

test_that("build_config_list: headerless writes col_names but no csv_skip", {
  wiz <- make_wiz(
    file_mode  = "explicit", current_file = "/data/nohdr.csv",
    has_header = FALSE,
    col_names  = c("a", "b", "c")
  )
  cfg <- build_config_list(wiz)
  expect_equal(unlist(cfg$col_names), c("a", "b", "c"))
  expect_null(cfg$csv_skip)
})

test_that("renamed-header config round-trips through write -> read -> rebuild (Gap 1)", {
  # The cycle that a single-save test hides: a renamed-header config must come
  # back as has_header = TRUE and re-serialise identically on the SECOND save.
  wiz1 <- make_wiz(
    dataset_name     = "refunds",
    file_mode        = "explicit", current_file = "/data/refunds.csv",
    has_header       = TRUE,
    raw_header_names = c("PayeeName", "PayeeName", "Amount"),
    col_names        = c("PayeeName", "PayeeName_2", "Amount")
  )
  path <- tempfile(fileext = ".yml")
  write_config(wiz1, list(), path)
  cfg1 <- yaml::read_yaml(path)
  expect_equal(cfg1$csv_skip, 1L)

  # Reopen (as the edit wizard would) — header presence must be restored.
  k <- read_config(path)$known
  expect_true(k$has_header)
  expect_equal(k$col_names, c("PayeeName", "PayeeName_2", "Amount"))

  # Reseed a wiz from the loaded config and re-serialise. On reopen the raw
  # header probe would repopulate raw_header_names from the file; simulate that.
  wiz2 <- make_wiz(
    dataset_name     = "refunds",
    file_mode        = "explicit", current_file = "/data/refunds.csv",
    has_header       = k$has_header,
    raw_header_names = c("PayeeName", "PayeeName", "Amount"),
    col_names        = k$col_names
  )
  cfg2 <- build_config_list(wiz2)
  expect_equal(unlist(cfg2$col_names), c("PayeeName", "PayeeName_2", "Amount"))
  expect_equal(cfg2$csv_skip, 1L)   # NOT dropped on the second save
})

# Step 3 CSV branch server logic (spec §11.3–11.4)

sniff_csv_file <- function(path) {
  tryCatch({
    lines <- readLines(path, n = 100, warn = FALSE)
    lines <- lines[nchar(trimws(lines)) > 0]
    if (length(lines) < 1) return(NULL)

    candidates <- c(",", "\t", ";", "|", ":")
    best_delim <- ","
    best_score <- 0
    for (d in candidates) {
      counts <- vapply(lines, function(l) length(strsplit(l, d, fixed = TRUE)[[1]]), integer(1))
      n_cols  <- stats::median(counts)
      if (n_cols > 1) {
        score <- (n_cols - 1) * mean(counts == n_cols)
        if (score > best_score) { best_score <- score; best_delim <- d }
      }
    }

    first_fields <- strsplit(lines[1], best_delim, fixed = TRUE)[[1]]
    has_header <- all(is.na(suppressWarnings(as.numeric(trimws(first_fields)))))

    df <- readr::read_delim(path, delim = best_delim,
                            col_names = has_header,
                            col_types = readr::cols(.default = "c"),
                            n_max = 500, show_col_types = FALSE)

    list(delimiter  = best_delim,
         has_header = has_header,
         col_names  = names(df),
         col_types  = vapply(df, infer_col_type_simple, character(1)))
  }, error = function(e) NULL)
}

load_raw_preview <- function(session, path, wiz) {
  lines <- tryCatch(
    readLines(path, n=50, warn=FALSE, encoding="bytes"),
    error=function(e) paste("Could not read file:", e$message)
  )
  wiz$raw_lines <- lines
  shinyAce::updateAceEditor(session, "raw_preview",
    value=paste(lines, collapse="\n"))
  # Update ruler
  max_chars <- max(nchar(lines), na.rm=TRUE)
  max_chars <- min(max(max_chars, 40), 200)
  ruler_str <- make_ruler_string(max_chars)
  wiz$ruler_string <- ruler_str
}

server_step3_csv <- function(input, output, session, wiz) {

  # CSV fields shown/hidden
  output$step3_csv_fields <- renderUI({
    req(input$wiz_format == "csv")
    tagList(
      fluidRow(
        column(3,
          selectInput("wiz_delimiter", "Delimiter",
            choices=c("Comma (,)"=",", "Tab"="\t", "Semicolon (;)"=";",
                      "Pipe (|)"="|", "Space"=" ", "Colon (:)"=":",
                      "Other"="other"),
            selected=wiz$delimiter %||% ",")
        ),
        column(3,
          uiOutput("step3_other_delimiter")
        ),
        column(3,
          selectInput("wiz_encoding", "Encoding",
            choices=wiz$encoding_choices %||% c("UTF-8","ISO-8859-1","Windows-1252","UTF-16LE","CP1250"),
            selected=wiz$encoding %||% "UTF-8")
        ),
        column(3,
          selectInput("wiz_quote_char", "Quote character",
            choices=c('Double quote "'='"', "Single quote '"="'", "None"=""),
            selected=wiz$quote_char %||% '"')
        )
      ),
      fluidRow(
        column(6,
          radioButtons("wiz_has_header", "First row",
            choices=c("Contains column headers"=TRUE, "Is data — no headers"=FALSE),
            selected=as.character(isTRUE(wiz$has_header %||% TRUE)))
        )
      ),
      div(class="mt-2",
        actionButton("step3_preview_btn", "Preview with these settings",
                     class="btn btn-outline-primary btn-sm")
      )
    )
  })

  output$step3_other_delimiter <- renderUI({
    req(input$wiz_delimiter == "other")
    textInput("wiz_delimiter_custom", "Custom:", value="", width="80px")
  })

  output$step3_fwf_fields <- renderUI({
    req(input$wiz_format == "fwf")
    tagList(
      fluidRow(
        column(4,
          numericInput("wiz_fwf_skip", "Header rows to skip",
                       value=wiz$fwf_skip %||% 0L, min=0, max=10, step=1)
        ),
        column(8,
          actionButton("step3_fwf_autodetect", "Auto-detect boundaries",
                       class="btn btn-outline-secondary btn-sm mt-4"),
          p(class="text-muted mt-1", style="font-size:11px;",
            "Works best when columns are space-separated. For tightly packed files, click on the text above to place boundaries.")
        )
      ),
      uiOutput("fwf_col_def_table"),
      uiOutput("fwf_record_length_badge")
    )
  })

  # Format hint below raw preview
  output$step3_format_hint <- renderUI({
    if (isTRUE(input$wiz_format == "fwf")) {
      div(class="alert alert-info p-2 mt-1", style="font-size:12px;",
          "Click on the text above to add a column boundary. Drag a boundary to move it. Double-click a boundary to remove it.",
          br(),
          "The character position is shown in the bottom-right of the preview.")
    }
  })

  # Auto-run sniff when file loaded.
  # New mode: auto-apply sniffed values (no config to preserve).
  # Edit mode: compare sniffed vs config-loaded; surface conflicts at step 3 without
  # overwriting. The user decides per-field whether to keep the config value or adopt
  # the detected one.
  observe({
    req(length(wiz$raw_lines) > 0)
    path <- wiz$current_preview_path
    req(nchar(path %||% "") > 0)

    sniff <- sniff_csv_file(path)

    if (!is.null(sniff) && !is.null(sniff$col_names)) {
      wiz$sniff_col_names <- sniff$col_names
      wiz$sniff_col_types <- sniff$col_types
    }

    enc_raw <- tryCatch(readr::guess_encoding(path), error = function(e) NULL)
    top_enc     <- NULL
    enc_choices <- NULL
    if (!is.null(enc_raw) && nrow(enc_raw) > 0) {
      top_enc <- enc_raw$encoding[1]
      enc_choices <- unique(c(
        setNames(enc_raw$encoding[seq_len(min(3, nrow(enc_raw)))],
                 sprintf("%s (%.0f%%)", enc_raw$encoding[seq_len(min(3, nrow(enc_raw)))],
                         enc_raw$confidence[seq_len(min(3, nrow(enc_raw)))] * 100)),
        c("UTF-8" = "UTF-8", "ISO-8859-1" = "ISO-8859-1",
          "Windows-1252" = "Windows-1252", "UTF-16LE" = "UTF-16LE", "CP1250" = "CP1250")
      ))
      wiz$encoding_choices <- enc_choices
    }

    if (isTRUE(wiz$mode == "edit")) {
      conflicts <- list()

      if (!is.null(sniff) && isTRUE(wiz$format == "csv")) {
        detected_delim <- sniff$delimiter %||% ","
        if (detected_delim %in% c(",", "\t", ";", "|", " ", ":") &&
            detected_delim != (wiz$delimiter %||% ",")) {
          conflicts[["delimiter"]] <- list(
            field       = "delimiter",
            config_val  = wiz$delimiter %||% ",",
            sniffed_val = detected_delim,
            label       = "Delimiter"
          )
        }
        sniff_hdr <- isTRUE(sniff$has_header)
        if (sniff_hdr != isTRUE(wiz$has_header)) {
          conflicts[["has_header"]] <- list(
            field       = "has_header",
            config_val  = isTRUE(wiz$has_header),
            sniffed_val = sniff_hdr,
            label       = "Header row"
          )
        }
      }

      if (!is.null(top_enc) && top_enc != (wiz$encoding %||% "UTF-8")) {
        conflicts[["encoding"]] <- list(
          field       = "encoding",
          config_val  = wiz$encoding %||% "UTF-8",
          sniffed_val = top_enc,
          label       = "Encoding"
        )
      }

      if (!is.null(enc_choices))
        updateSelectInput(session, "wiz_encoding",
                          choices  = enc_choices,
                          selected = wiz$encoding %||% "UTF-8")
      wiz$sniff_conflicts <- conflicts

    } else {
      if (!is.null(sniff)) {
        detected_delim <- sniff$delimiter %||% ","
        if (detected_delim %in% c(",", "\t", ";", "|", " ", ":")) {
          updateSelectInput(session, "wiz_delimiter", selected = detected_delim)
          wiz$delimiter <- detected_delim
        }
        has_hdr <- isTRUE(sniff$has_header)
        updateRadioButtons(session, "wiz_has_header", selected = as.character(has_hdr))
        wiz$has_header <- has_hdr
      }
      if (!is.null(enc_choices) && !is.null(top_enc)) {
        updateSelectInput(session, "wiz_encoding", choices = enc_choices, selected = top_enc)
        wiz$encoding <- top_enc
      }
    }
  })

  # Conflict banner — shown at step 3 in edit mode when sniffed values differ from config
  output$step3_sniff_conflicts <- renderUI({
    conflicts <- wiz$sniff_conflicts
    if (length(conflicts) == 0) return(NULL)

    fmt_val <- function(field, val) {
      if (field == "has_header") {
        if (isTRUE(val)) "has header row" else "no header row"
      } else {
        as.character(val)
      }
    }

    panels <- lapply(conflicts, function(cf) {
      div(class = "alert alert-warning p-2 mb-2", style = "font-size:12px;",
        tags$strong(paste0(cf$label, " mismatch — ")),
        sprintf("config says \"%s\", file looks like \"%s\".",
                fmt_val(cf$field, cf$config_val),
                fmt_val(cf$field, cf$sniffed_val)),
        div(class = "mt-1 d-flex gap-2",
          actionButton(paste0("sniff_keep_", cf$field),
                       paste0("Keep config: ", fmt_val(cf$field, cf$config_val)),
                       class = "btn btn-outline-secondary btn-sm",
                       style = "font-size:11px; padding:2px 8px;"),
          actionButton(paste0("sniff_use_", cf$field),
                       paste0("Use detected: ", fmt_val(cf$field, cf$sniffed_val)),
                       class = "btn btn-outline-primary btn-sm",
                       style = "font-size:11px; padding:2px 8px;")
        )
      )
    })
    tagList(panels)
  })

  # Dismiss conflict — keep config value
  observeEvent(input$sniff_keep_delimiter, {
    cf <- wiz$sniff_conflicts; cf[["delimiter"]] <- NULL; wiz$sniff_conflicts <- cf
  })
  observeEvent(input$sniff_keep_has_header, {
    cf <- wiz$sniff_conflicts; cf[["has_header"]] <- NULL; wiz$sniff_conflicts <- cf
  })
  observeEvent(input$sniff_keep_encoding, {
    cf <- wiz$sniff_conflicts; cf[["encoding"]] <- NULL; wiz$sniff_conflicts <- cf
  })

  # Accept detected value and dismiss
  observeEvent(input$sniff_use_delimiter, {
    cf <- wiz$sniff_conflicts[["delimiter"]]; req(!is.null(cf))
    wiz$delimiter <- cf$sniffed_val
    updateSelectInput(session, "wiz_delimiter", selected = cf$sniffed_val)
    cfs <- wiz$sniff_conflicts; cfs[["delimiter"]] <- NULL; wiz$sniff_conflicts <- cfs
  })
  observeEvent(input$sniff_use_has_header, {
    cf <- wiz$sniff_conflicts[["has_header"]]; req(!is.null(cf))
    v <- isTRUE(cf$sniffed_val)
    wiz$has_header <- v
    updateRadioButtons(session, "wiz_has_header", selected = as.character(v))
    cfs <- wiz$sniff_conflicts; cfs[["has_header"]] <- NULL; wiz$sniff_conflicts <- cfs
  })
  observeEvent(input$sniff_use_encoding, {
    cf <- wiz$sniff_conflicts[["encoding"]]; req(!is.null(cf))
    wiz$encoding <- cf$sniffed_val
    updateSelectInput(session, "wiz_encoding", selected = cf$sniffed_val)
    cfs <- wiz$sniff_conflicts; cfs[["encoding"]] <- NULL; wiz$sniff_conflicts <- cfs
  })

  # Update wiz from inputs
  observeEvent(input$wiz_format,    { wiz$format    <- input$wiz_format
    if (input$wiz_format == "fwf") session$sendCustomMessage("fwf_ruler_activate", list())
    else session$sendCustomMessage("fwf_ruler_deactivate", list()) })
  observeEvent(input$wiz_delimiter, { wiz$delimiter  <- if(input$wiz_delimiter=="other") input$wiz_delimiter_custom %||% "," else input$wiz_delimiter })
  observeEvent(input$wiz_encoding,  { wiz$encoding   <- input$wiz_encoding })
  observeEvent(input$wiz_quote_char,{ wiz$quote_char <- input$wiz_quote_char })
  observeEvent(input$wiz_has_header,{ wiz$has_header <- isTRUE(as.logical(input$wiz_has_header)) })
  observeEvent(input$wiz_fwf_skip,  { wiz$fwf_skip   <- as.integer(input$wiz_fwf_skip %||% 0) })

  # Preview trigger (CSV)
  observeEvent(input$step3_preview_btn, {
    trigger_csv_preview(input, wiz)
  })
  # Also auto-preview when format is CSV and file is loaded
  observe({
    req(input$wiz_format == "csv", length(wiz$raw_lines) > 0)
    trigger_csv_preview(input, wiz)
  })

  # FWF boundary positions → widths
  observeEvent(input$fwf_boundary_positions, {
    positions <- sort(unique(c(0L, as.integer(input$fwf_boundary_positions))))
    line_len  <- max(nchar(wiz$raw_lines[nchar(wiz$raw_lines)>0]), na.rm=TRUE)
    if (!is.finite(line_len)) line_len <- 80L
    widths <- diff(c(positions, as.integer(line_len)))
    widths <- widths[widths > 0]
    wiz$fwf_widths <- widths
    wiz$fwf_starts <- positions[positions > 0] + 1L
    wiz$fwf_line_len <- line_len
  })

  # FWF auto-detect
  observeEvent(input$step3_fwf_autodetect, {
    path <- wiz$current_preview_path
    req(nchar(path %||% "") > 0)
    result <- tryCatch(
      readr::fwf_empty(path, skip=as.integer(input$wiz_fwf_skip %||% 0), n=100),
      error=function(e) NULL
    )
    if (!is.null(result) && nrow(result) > 1) {
      # Convert to 0-based char positions (boundary starts)
      positions <- result$begin[-1]  # skip the first (=0)
      session$sendCustomMessage("fwf_restore_boundaries", list(positions=as.list(positions)))
      showNotification(sprintf("Auto-detected %d columns — please verify.", nrow(result)), type="message")
    } else {
      showNotification("Auto-detection found no boundaries. Set manually using the ruler.", type="warning")
    }
  })

  # FWF column name inputs
  output$fwf_col_def_table <- renderUI({
    n <- length(wiz$fwf_widths)
    if (n == 0) return(p(class="text-muted fst-italic", "No column boundaries set yet. Click on the text preview above."))

    starts <- c(1L, cumsum(wiz$fwf_widths[-length(wiz$fwf_widths)]) + 1L)

    rows <- lapply(seq_len(n), function(i) {
      name_id <- paste0("fwf_col_name_", i)
      type_id <- paste0("fwf_col_type_", i)
      current_name <- if (i <= length(wiz$fwf_col_names)) wiz$fwf_col_names[i] else paste0("col_", i)
      fluidRow(class="mb-1 align-items-center",
        column(1, span(class="badge bg-secondary", i)),
        column(2, span(class="text-muted", style="font-size:12px;", sprintf("%d–%d", starts[i], starts[i]+wiz$fwf_widths[i]-1L))),
        column(2, span(class="text-muted", style="font-size:12px;", paste("w:", wiz$fwf_widths[i]))),
        column(4, textInput(name_id, NULL, value=current_name, placeholder=paste0("col_", i))),
        column(3, selectInput(type_id, NULL, width="100%",
          choices=c("character","numeric","date"), selected="character"))
      )
    })

    tagList(
      div(class="mt-2 mb-1", tags$strong("Column definitions", style="font-size:13px;")),
      div(style="max-height:300px;overflow-y:auto;", rows)
    )
  })

  # Collect FWF column names from inputs
  observe({
    n <- length(wiz$fwf_widths)
    if (n == 0) return()
    names_vec <- character(n)
    for (i in seq_len(n)) {
      v <- input[[paste0("fwf_col_name_", i)]]
      names_vec[i] <- if (!is.null(v) && nchar(v) > 0) v else paste0("col_", i)
    }
    wiz$fwf_col_names <- names_vec
  })

  # Collect FWF column types from inputs
  observe({
    n <- length(wiz$fwf_widths)
    if (n == 0) return()
    col_names <- wiz$fwf_col_names
    types_vec <- character(n)
    for (i in seq_len(n)) {
      v <- input[[paste0("fwf_col_type_", i)]]
      types_vec[i] <- if (!is.null(v)) v else "character"
    }
    names(types_vec) <- if (length(col_names) == n) col_names else paste0("col_", seq_len(n))
    wiz$col_types_inferred <- types_vec
  })

  # FWF record length badge
  output$fwf_record_length_badge <- renderUI({
    req(length(wiz$fwf_widths) > 0, length(wiz$raw_lines) > 0)
    line_len    <- max(nchar(wiz$raw_lines[nchar(wiz$raw_lines) > 0]), na.rm=TRUE)
    total_width <- sum(wiz$fwf_widths)
    if (!is.finite(line_len)) return(NULL)

    if (total_width == line_len) {
      div(class="alert alert-success p-1 mt-2", style="font-size:12px;",
          sprintf("✅ Record length: %d / %d characters", total_width, line_len))
    } else if (total_width < line_len) {
      div(class="alert alert-warning p-1 mt-2", style="font-size:12px;",
          sprintf("⚠ Column definitions cover %d of %d characters — %d characters unaccounted for at end of record.",
                  total_width, line_len, line_len - total_width))
    } else {
      div(class="alert alert-danger p-1 mt-2", style="font-size:12px;",
          sprintf("✗ Column definitions (%d) exceed actual line length (%d).", total_width, line_len))
    }
  })

  # Column-naming panel. Renders for headerless files (always) and for header
  # files whose raw header names are invalid or duplicated (e.g. RBB RefundsPaid
  # repeats `PayeeName`). Pre-fills suggestions with an inline reason; fully
  # overridable. Input ids are shared across both modes (`csv_name_<i>`).
  output$step3_no_header_naming <- renderUI({
    req(input$wiz_format == "csv")
    has_header <- isTRUE(as.logical(input$wiz_has_header %||% "TRUE"))
    cols <- wiz$csv_col_names_detected          # reactive trigger: the column SET
    if (is.null(cols) || length(cols) == 0) return(NULL)
    if (!csv_needs_naming(wiz, has_header)) return(NULL)

    # Read the editable values WITHOUT subscribing to them. The collector below
    # writes wiz$col_names on every keystroke; if this panel re-rendered on each
    # change it would rebuild the DOM mid-edit, resetting the preview's scroll
    # position and fighting the cursor. We only want to rebuild when the column
    # set itself changes (handled by the reactive reads above). Per CLAUDE.md,
    # isolate() is the wizard's standard tool for exactly this.
    isolate({
      reasons <- wiz$col_name_reasons
      rows <- lapply(seq_along(cols), function(i) {
        name_id <- paste0("csv_name_", i)
        current <- if (i <= length(wiz$col_names)) wiz$col_names[i] else paste0("col_", i)
        left    <- if (has_header) cols[i] else paste0("col_", i)
        hint    <- if (length(reasons) >= i && nzchar(reasons[i]))
          span(class="text-warning", style="font-size:11px;", paste0("⚠ ", reasons[i]))
        else NULL
        fluidRow(class="mb-1 align-items-center",
          column(3, span(class="text-muted", style="font-size:12px;", paste0(left, " →"))),
          column(4, textInput(name_id, NULL, value=current, placeholder=paste0("col_",i), width="100%")),
          column(5, hint)
        )
      })

      div(class="p-2 border rounded bg-light mb-3",
        tags$strong("Column names", style="font-size:13px;"),
        if (has_header)
          p(class="text-muted mb-1", style="font-size:11px;",
            "Some names in the file's header are invalid or duplicated. Adjust as needed — the original header row will be skipped when the file is read."),
        div(style="max-height:240px;overflow-y:auto;", rows)
      )
    })
  })

  # Collect names from the editor (runs under the same condition that renders it)
  observe({
    req(input$wiz_format == "csv")
    has_header <- isTRUE(as.logical(input$wiz_has_header %||% "TRUE"))
    cols <- wiz$csv_col_names_detected
    req(!is.null(cols), length(cols) > 0)
    if (!csv_needs_naming(wiz, has_header)) return()
    names_vec <- character(length(cols))
    for (i in seq_along(cols)) {
      v <- input[[paste0("csv_name_", i)]]
      names_vec[i] <- if (!is.null(v) && nchar(v) > 0) v
        else if (i <= length(wiz$col_names)) wiz$col_names[i]
        else paste0("col_", i)
    }
    wiz$col_names <- names_vec
  })

  # Parsed preview (CSV)
  output$step3_parsed_preview <- renderUI({
    req(input$wiz_format == "csv", !is.null(wiz$csv_preview_df))
    reactable::reactableOutput("step3_reactable_preview")
  })

  output$step3_reactable_preview <- reactable::renderReactable({
    req(!is.null(wiz$csv_preview_df))
    df <- wiz$csv_preview_df
    reactable::reactable(df,
      resizable=TRUE, wrap=FALSE, fullWidth=TRUE,
      defaultPageSize=20, striped=TRUE, highlight=TRUE,
      style=list(fontSize="12px"))
  })

  # FWF parsed preview
  observe({
    req(input$wiz_format == "fwf",
        length(wiz$fwf_widths) > 0,
        length(wiz$fwf_col_names) > 0,
        nchar(wiz$current_preview_path %||% "") > 0)

    path <- wiz$current_preview_path
    tryCatch({
      df <- readr::read_fwf(path,
        col_positions=readr::fwf_widths(wiz$fwf_widths, wiz$fwf_col_names),
        col_types=readr::cols(.default="c"),
        n_max=20,
        skip=as.integer(wiz$fwf_skip %||% 0),
        locale=readr::locale(encoding=wiz$encoding %||% "UTF-8"))
      wiz$fwf_preview_df <- as.data.frame(df)
    }, error=function(e) { wiz$fwf_preview_df <- NULL })
  })

  # FWF parsed preview output
  output$step3_fwf_preview <- reactable::renderReactable({
    req(!is.null(wiz$fwf_preview_df))
    reactable::reactable(wiz$fwf_preview_df,
      resizable=TRUE, wrap=FALSE, fullWidth=TRUE,
      defaultPageSize=20, striped=TRUE, highlight=TRUE,
      style=list(fontSize="12px"))
  })

}

# Helper to trigger CSV re-parse
trigger_csv_preview <- function(input, wiz) {
  path <- wiz$current_preview_path
  if (is.null(path) || nchar(path) == 0) return()

  delim <- if (!is.null(input$wiz_delimiter) && input$wiz_delimiter == "other")
    input$wiz_delimiter_custom %||% ","
  else input$wiz_delimiter %||% ","
  enc  <- input$wiz_encoding %||% "UTF-8"
  hdr  <- isTRUE(as.logical(input$wiz_has_header %||% "TRUE"))
  quot <- input$wiz_quote_char %||% '"'

  tryCatch({
    df <- readr::read_delim(path,
      delim=delim,
      col_names=hdr,
      col_types=readr::cols(.default="c"),
      n_max=20,
      quote=quot,
      locale=readr::locale(encoding=enc),
      show_col_types=FALSE)
    df <- as.data.frame(df)
    ncols <- ncol(df)

    # `wiz$col_names` is the single source of truth for column names. Reseed it
    # ONLY when the column count changes — never clobber an existing list of the
    # right length, or we would erase the user's edits / our own suggestions on
    # every re-preview (the auto-preview observer fires on each step-3 render).
    if (!hdr) {
      wiz$raw_header_names <- character(0)
      wiz$col_name_reasons <- character(0)
      wiz$csv_col_names_detected <- names(df)
      if (length(wiz$col_names) != ncols)
        wiz$col_names <- paste0("col_", seq_len(ncols))
    } else {
      # Probe the *unmangled* header so duplicates survive (readr's default
      # name_repair mangles them to `name...NN`, which we can't fix in the UI).
      raw <- tryCatch(
        names(readr::read_delim(path, delim=delim, n_max=0, quote=quot,
                                name_repair="minimal",
                                locale=readr::locale(encoding=enc),
                                show_col_types=FALSE)),
        error=function(e) names(df))
      if (length(raw) != ncols) raw <- names(df)   # safety: keep counts aligned
      sug <- suggest_col_names(raw)
      wiz$raw_header_names <- raw
      wiz$col_name_reasons <- sug$reason
      wiz$csv_col_names_detected <- raw
      if (length(wiz$col_names) != ncols)
        wiz$col_names <- sug$names
    }

    # Show the preview under the resolved names, not the raw/mangled ones.
    if (length(wiz$col_names) == ncols) names(df) <- wiz$col_names

    # Infer types from sample
    wiz$col_types_inferred <- vapply(df, function(x) infer_col_type_simple(x), character(1))
    wiz$csv_preview_df <- df
  }, error=function(e) {
    wiz$csv_preview_df <- data.frame(Error=paste("Parse failed:", e$message))
  })
}

# Run panel UI (spec §17)

ui_run <- function(datasets, selected = NULL) {
  tagList(
    h4("Run Quality Check", style="margin-bottom:20px;"),

    fluidRow(
      column(6,
        selectInput("run_dataset", "Dataset", choices=datasets,
                    selected=selected, width="100%")
      )
    ),

    uiOutput("run_precheck_status"),

    div(class="mt-3 d-flex gap-2 align-items-center",
      uiOutput("run_start_btn"),
      uiOutput("run_stop_btn")
    ),

    div(class="mt-3",
      uiOutput("run_status_area")
    ),

    div(class="mt-2",
      uiOutput("run_log_area")
    )
  )
}

# Run panel server logic (spec §17)

server_run <- function(input, output, session, rv, config_dir, gcfg_rv) {

  rv_run <- reactiveValues(
    r_process  = NULL,
    log_path   = NULL,
    log_lines  = character(0),
    status     = "idle",
    report_path = NULL,
    error_msg  = NULL
  )

  # Update dataset selector
  observe({
    datasets <- list_dataset_configs(config_dir())
    selected <- if (!is.null(rv$run_preselect) && rv$run_preselect %in% datasets)
                  rv$run_preselect else datasets[1]
    updateSelectInput(session, "run_dataset",
                      choices=if(length(datasets)>0) datasets else list("(no datasets)"=""),
                      selected=selected)
  })

  # Pre-run validation
  output$run_precheck_status <- renderUI({
    ds <- input$run_dataset
    if (is.null(ds) || nchar(ds) == 0)
      return(div(class="alert alert-warning p-2 mb-2", style="font-size:12px;",
                 "Select a dataset to run."))
    cd <- config_dir()
    cfg_path <- file.path(cd, paste0(ds, ".yml"))
    gcfg <- gcfg_rv()

    issues <- character(0)
    if (!safe_file_exists(cfg_path))
      issues <- c(issues, paste("Config file not found:", cfg_path))

    if (safe_file_exists(cfg_path)) {
      cfg <- tryCatch(read_config(cfg_path)$known, error=function(e) NULL)
      if (!is.null(cfg)) {
        # Check file/folder exists
        if (nchar(cfg$folder %||% "") > 0 && !safe_dir_exists(cfg$folder))
          issues <- c(issues, paste("Folder not found:", cfg$folder))
        if (nchar(cfg$current_file %||% "") > 0 && !safe_file_exists(cfg$current_file))
          issues <- c(issues, paste("Current file not found:", cfg$current_file))
      }
    }

    db_path <- resolve_infra_path(gcfg$snapshot_db, cd,
                                  default = "data/snapshots.sqlite", mustWork = FALSE)
    if (nchar(db_path) > 0) {
      db_parent <- dirname(db_path)
      if (!safe_dir_exists(db_parent))
        issues <- c(issues, paste("Snapshot DB directory not found:", db_parent))
    }

    if (length(issues) == 0) {
      div(class="alert alert-success p-2 mb-2", style="font-size:12px;",
          "✓ Configuration looks good.")
    } else {
      div(class="alert alert-danger p-2 mb-2", style="font-size:12px;",
          tags$strong("Issues found:"),
          tags$ul(lapply(issues, tags$li)))
    }
  })

  # Run button — disabled while a run is in progress
  output$run_start_btn <- renderUI({
    ds  <- input$run_dataset
    cd  <- config_dir()
    ok  <- !is.null(ds) && nchar(ds) > 0 &&
           safe_file_exists(file.path(cd, paste0(ds, ".yml"))) &&
           rv_run$status != "running"
    btn <- actionButton("run_start", "▶ Run check", class = "btn btn-primary")
    if (!ok) tagAppendAttributes(btn, disabled = "disabled") else btn
  })

  # Run start
  observeEvent(input$run_start, {
    req(input$run_dataset, rv_run$status != "running")
    ds <- input$run_dataset
    cd <- config_dir()

    rv_run$log_path  <- tempfile("dqcheckr_run_", fileext=".log")
    rv_run$log_lines <- character(0)
    rv_run$status    <- "running"
    rv_run$report_path <- NULL
    rv_run$error_msg   <- NULL

    rv_run$r_process <- callr::r_bg(
      func   = function(dn, cd) dqcheckr::run_dq_check(dn, config_dir=cd),
      args   = list(dn=ds, cd=cd),
      stdout = rv_run$log_path,
      stderr = "2>&1",
      # Run in the deployment root so dqcheckr resolves relative snapshot_db /
      # report_output_dir the same way the CLI does (shiny::runApp has changed
      # this process's getwd() to the installed app directory).
      wd      = deployment_root(cd),
      package = TRUE
    )
  })

  # 200ms polling observer
  observe({
    req(rv_run$status == "running")
    invalidateLater(200)

    if (safe_file_exists(rv_run$log_path)) {
      rv_run$log_lines <- readLines(rv_run$log_path, warn=FALSE)
      session$sendCustomMessage("scroll_log", list())
    }

    proc <- rv_run$r_process
    if (!is.null(proc) && !proc$is_alive()) {
      # On failure, surface the *actual* cause — either the condition message
      # from the background process, or (more often informative) the last
      # non-blank line of its log — rather than a generic "check the log"
      # message. Mirrors the pattern already used for drift failures in
      # server_history.R's poll observer.
      result <- tryCatch(proc$get_result(), error = function(e) {
        last_line <- tryCatch({
          lp <- rv_run$log_path
          if (!is.null(lp) && file.exists(lp)) {
            lines <- readLines(lp, warn = FALSE)
            lines <- lines[nchar(trimws(lines)) > 0]
            if (length(lines) > 0) tail(lines, 1) else ""
          } else ""
        }, error = function(e2) "")
        msg <- if (nchar(last_line) > 0) last_line else conditionMessage(e)
        rv_run$error_msg <- paste("Run failed:", msg)
        NULL
      })
      rv_run$status <- if (!is.null(result)) as.character(result$status) else "error"
      if (!is.null(result)) rv_run$report_path <- result$report_path
      rv$history_refresh <- Sys.time()
    }
  })

  # Stop button
  output$run_stop_btn <- renderUI({
    if (rv_run$status == "running")
      actionButton("run_stop", "■ Stop", class="btn btn-danger btn-sm")
  })

  observeEvent(input$run_stop, {
    showModal(modalDialog(
      title="Stop run?",
      "Stop the current run? The snapshot will not be written if the run has not yet completed.",
      footer=tagList(
        modalButton("Keep running"),
        actionButton("run_stop_confirm", "Stop", class="btn btn-danger")
      )
    ))
  })

  observeEvent(input$run_stop_confirm, {
    removeModal()
    if (!is.null(rv_run$r_process)) {
      tryCatch(rv_run$r_process$kill(), error=function(e) NULL)
    }
    rv_run$status <- "stopped"
    rv_run$log_lines <- c(rv_run$log_lines, "\n[Run stopped by user]")
  })

  # Status area
  output$run_status_area <- renderUI({
    s <- rv_run$status
    if (s == "idle") return(NULL)
    div(class="d-flex gap-3 align-items-center mt-2",
      status_badge(toupper(s)),
      if (s %in% c("PASS","WARN","FAIL") && !is.null(rv_run$report_path)) {
        report_url <- paste0("/dq_reports/", basename(rv_run$report_path))
        div(
          tags$a(href=report_url, target="_blank",
                 class="btn btn-outline-primary btn-sm", "Open report ↗"),
          actionButton("run_view_log", "View log",
                       class="btn btn-outline-secondary btn-sm ms-2")
        )
      } else if (s == "stopped") {
        span(class="text-muted fst-italic", "Run stopped by user.")
      } else if (s == "error") {
        span(class="text-danger", rv_run$error_msg %||% "Run failed.")
      }
    )
  })

  # Log output
  output$run_log_area <- renderUI({
    if (rv_run$status == "idle") return(NULL)
    div(
      verbatimTextOutput("run_log"),
    )
  })

  output$run_log <- renderText({
    req(length(rv_run$log_lines) > 0)
    paste(rv_run$log_lines, collapse="\n")
  })

  observeEvent(input$run_view_log, {
    req(rv_run$log_path, safe_file_exists(rv_run$log_path))
    log_text <- paste(readLines(rv_run$log_path, warn=FALSE), collapse="\n")
    showModal(modalDialog(
      title="Run log",
      pre(log_text, style="font-size:11px;max-height:500px;overflow-y:auto;"),
      easyClose=TRUE, size="xl"
    ))
  })
}

#' Launch the dqcheckrGUI Shiny application
#'
#' Opens the point-and-click interface for configuring and running
#' \pkg{dqcheckr} dataset quality checks.
#'
#' @param ... Arguments passed to \code{\link[shiny]{runApp}} (e.g.
#'   \code{port}, \code{launch.browser}).
#'
#' @return Called for its side effect; does not return a value.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   run_app()
#' }
run_app <- function(...) {
  app_dir <- system.file("app", package = "dqcheckrGUI")
  if (app_dir == "") {
    stop("Could not find app directory. Try re-installing dqcheckrGUI.")
  }
  shiny::runApp(app_dir, ...)
}

# History panel UI (spec §18)

ui_history <- function() {
  tagList(
    h4("Run History", style="margin-bottom:20px;"),
    div(class="d-flex gap-2 align-items-center mb-3",
      actionButton("history_compare", "Compare drift ▶",
                   class="btn btn-outline-info btn-sm", disabled=TRUE),
      span(class="text-muted fst-italic", style="font-size:12px;",
           "Select exactly 2 runs from the same dataset")
    ),
    DT::dataTableOutput("history_table"),
    div(class="mt-2",
      actionButton("history_load_more", "Load more", class="btn btn-outline-secondary btn-sm")
    )
  )
}

# Global config editor UI (spec §19)

ui_global_config <- function(gcfg) {
  dr <- gcfg$default_rules %||% list()

  tagList(
    h4("Global Configuration", style="margin-bottom:20px;"),

    bslib::card(
      bslib::card_header("Infrastructure Paths"),
      bslib::card_body(
        fluidRow(
          column(8,
            tags$label("Snapshot database", class="form-label fw-semibold"),
            div(class="d-flex gap-2",
              textInput("gcfg_snapshot_db", NULL,
                        value = gcfg$snapshot_db %||% "data/snapshots.sqlite",
                        width = "100%"),
              shinyFiles::shinyFilesButton("gcfg_db_browse", "Browse",
                "Select snapshot database file", FALSE,
                style="white-space:nowrap;height:38px;")
            ),
            uiOutput("sv_gcfg_snapshot_db")
          )
        ),
        fluidRow(
          column(8,
            tags$label("Report output directory", class="form-label fw-semibold mt-2"),
            div(class="d-flex gap-2",
              textInput("gcfg_report_dir", NULL,
                        value = gcfg$report_output_dir %||% "reports/",
                        width = "100%"),
              shinyFiles::shinyDirButton("gcfg_dir_browse", "Browse",
                "Select report output directory",
                style="white-space:nowrap;height:38px;")
            ),
            uiOutput("sv_gcfg_report_dir")
          )
        )
      )
    ),

    bslib::card(
      class = "mt-3",
      bslib::card_header("Default Rule Thresholds"),
      bslib::card_body(
        h6("Single-snapshot thresholds", class="text-muted mt-1"),
        fluidRow(
          column(4, numericInput("gcfg_max_missing_rate", "Max missing rate", value=dr$max_missing_rate %||% 0.05, min=0, max=1, step=0.01)),
          column(4, numericInput("gcfg_max_non_numeric_rate", "Max non-numeric rate", value=dr$max_non_numeric_rate %||% 0.01, min=0, max=1, step=0.01)),
          column(4, numericInput("gcfg_min_row_count", "Min row count (0=disabled)", value=dr$min_row_count %||% 0, min=0, step=1))
        ),
        h6("Version comparison thresholds", class="text-muted mt-3"),
        fluidRow(
          column(4, numericInput("gcfg_max_row_count_chg", "Max row count change (%)", value=round((dr$max_row_count_change_pct %||% 0.10)*100,2), min=0, max=100, step=1)),
          column(4, numericInput("gcfg_max_mean_shift", "Max mean shift (%)", value=round((dr$max_numeric_mean_shift_pct %||% 0.20)*100,2), min=0, max=100, step=1)),
          column(4, numericInput("gcfg_max_missing_chg", "Max missing rate change (pp)", value=dr$max_missing_rate_change_pp %||% 2.0, min=0, step=0.1))
        ),
        fluidRow(
          column(4, numericInput("gcfg_max_nonnumeric_chg", "Max non-numeric change (pp)", value=dr$max_non_numeric_rate_change_pp %||% 1.0, min=0, step=0.1)),
          column(4, numericInput("gcfg_type_inf_threshold", "Type inference threshold", value=dr$type_inference_threshold %||% 0.90, min=0, max=1, step=0.01))
        ),
        h6("Schema change flags", class="text-muted mt-3"),
        fluidRow(
          column(3, checkboxInput("gcfg_flag_new_cols",   "Flag new columns",   value=isTRUE(dr$flag_new_columns %||% TRUE))),
          column(3, checkboxInput("gcfg_flag_drop_cols",  "Flag dropped columns",value=isTRUE(dr$flag_dropped_columns %||% TRUE))),
          column(3, checkboxInput("gcfg_flag_type_chg",   "Flag type changes",   value=isTRUE(dr$flag_type_changes %||% TRUE))),
          column(3, checkboxInput("gcfg_flag_col_order",  "Flag column order",   value=isTRUE(dr$flag_column_order_change %||% TRUE)))
        )
      )
    ),

    div(class="mt-3",
      actionButton("gcfg_save", "Save global config", class="btn btn-primary")
    )
  )
}

# Config wizard server logic — state management, navigation, steps 1-2, 4-8
# Step 3 CSV/FWF split into server_step3_csv.R

server_wizard <- function(input, output, session, rv, config_dir, gcfg_rv) {

  # ── Wizard state ─────────────────────────────────────────────────────
  # Single source of truth for defaults — used to seed reactiveValues and to
  # reset state when the wizard is opened for a new dataset.
  .wiz_defaults <- list(
    mode="new", dataset_name="", description="",
    file_mode="folder", folder="", current_file="", previous_file="",
    format="csv", encoding="UTF-8", delimiter=",", quote_char='"',
    has_header=TRUE, csv_skip=0L,
    col_names=character(0), col_types_inferred=character(0),
    raw_header_names=character(0), col_name_reasons=character(0),
    col_types_override=list(),
    key_columns=character(0), expected_columns=character(0),
    fwf_widths=integer(0), fwf_col_names=character(0), fwf_skip=0L,
    column_rules=list(), rule_overrides=list(),
    custom_checks_file="", snapshot_db="", report_output_dir="",
    extra_keys=list(),
    sniff_conflicts=list(),
    current_step=1L,
    step_valid=rep(FALSE, 8L),
    raw_lines=character(0),
    ruler_string=make_ruler_string(80),
    current_preview_path="",
    encoding_choices=c("UTF-8","ISO-8859-1","Windows-1252","UTF-16LE","CP1250"),
    sniff_col_names=character(0), sniff_col_types=character(0),
    csv_preview_df=NULL, csv_col_names_detected=character(0),
    fwf_preview_df=NULL, fwf_starts=integer(0), fwf_line_len=80L
  )
  wiz <- do.call(reactiveValues, .wiz_defaults)

  open_wizard <- function(mode, dataset_name=NULL) {
    if (mode == "edit" && !is.null(dataset_name)) {
      path <- file.path(config_dir(), paste0(dataset_name, ".yml"))
      if (safe_file_exists(path)) {
        cfg <- tryCatch(read_config(path), error=function(e) NULL)
        if (!is.null(cfg)) {
          k <- cfg$known
          for (nm in names(k)) wiz[[nm]] <- k[[nm]]
          wiz$extra_keys    <- cfg$extra
          wiz$mode          <- "edit"
          wiz$current_step  <- 1L
          # Trigger file preview
          preview_path <- if (nchar(k$folder %||%"") > 0) {
            files <- list.files(k$folder, full.names=TRUE)
            if (length(files) > 0) files[order(file.mtime(files), decreasing=TRUE)[1]] else ""
          } else k$current_file %||% ""
          if (nchar(preview_path) > 0) {
            wiz$current_preview_path <- preview_path
            load_raw_preview(session, preview_path, wiz)
          }
        }
      }
    } else {
      for (nm in names(.wiz_defaults)) wiz[[nm]] <- .wiz_defaults[[nm]]
    }
    rv$wizard_open <- TRUE
    rv$active_section <- "wizard"
  }

  # Watch for wizard open requests from parent (rv$wizard_request)
  observeEvent(rv$wizard_request, {
    req(!is.null(rv$wizard_request))
    open_wizard(rv$wizard_request$mode, rv$wizard_request$dataset)
  })

  # ── Breadcrumb ────────────────────────────────────────────────────────
  output$wizard_breadcrumb <- renderUI({
    wizard_breadcrumb(wiz$current_step, wiz$step_valid)
  })

  # ── Step content ─────────────────────────────────────────────────────
  # Only react to step navigation (wiz$current_step); isolate all wiz field
  # reads so that typing in a text input doesn't re-render the whole step
  # and steal keyboard focus from the active field.
  output$wizard_step_content <- renderUI({
    s <- wiz$current_step          # reactive: re-render on step change only
    isolate({                      # isolate: wiz field reads don't subscribe
      gcfg <- gcfg_rv()
      if      (s == 1) wizard_step1_ui(wiz)
      else if (s == 2) wizard_step2_ui(wiz)
      else if (s == 3) wizard_step3_ui()
      else if (s == 4) wizard_step4_ui()
      else if (s == 5) wizard_step5_ui()
      else if (s == 6) wizard_step6_ui(wiz, gcfg)
      else if (s == 7) wizard_step7_ui(wiz)
      else if (s == 8) wizard_step8_ui()
    })
  })

  # ── Step validation ───────────────────────────────────────────────────
  step_valid <- reactive({
    valid <- rep(FALSE, 8L)
    # Step 1
    nm <- input$wiz_dataset_name %||% ""
    valid[1] <- nchar(nm) > 0 && is_valid_r_name(nm)
    # Step 2
    path <- wiz$current_preview_path
    valid[2] <- nchar(path %||% "") > 0
    # Step 3
    if (isTRUE(input$wiz_format == "fwf")) {
      # FWF valid when: widths sum <= line len, at least 1 col, all names valid
      total <- sum(wiz$fwf_widths)
      line_len <- wiz$fwf_line_len %||% 80L
      valid[3] <- total > 0 && total <= line_len &&
                  length(wiz$fwf_col_names) > 0 &&
                  all(nchar(wiz$fwf_col_names) > 0) &&
                  all(is_valid_r_name(wiz$fwf_col_names)) &&
                  !any(duplicated(wiz$fwf_col_names))
    } else {
      valid[3] <- length(wiz$col_names) > 0 &&
                  all(nchar(wiz$col_names) > 0) &&
                  all(is_valid_r_name(wiz$col_names)) &&
                  !any(duplicated(wiz$col_names))
    }
    # Steps 4-8 are unlocked once step 3 is complete (col_names populated)
    step3_done <- isTRUE(valid[3])
    valid[4] <- step3_done
    valid[5] <- step3_done
    valid[6] <- step3_done
    valid[7] <- step3_done
    valid[8] <- step3_done
    valid
  })

  observe({ wiz$step_valid <- step_valid() })

  # ── Navigation ────────────────────────────────────────────────────────
  output$wizard_back_btn <- renderUI({
    if (wiz$current_step > 1)
      actionButton("wizard_back", "◄ Back", class="btn btn-outline-secondary btn-sm")
  })
  output$wizard_next_btn <- renderUI({
    if (wiz$current_step < 8) {
      disabled <- !isTRUE(wiz$step_valid[wiz$current_step])
      btn <- actionButton("wizard_next", "Next ►", class="btn btn-primary btn-sm")
      if (disabled) tagAppendAttributes(btn, disabled=NA) else btn
    }
  })

  # (Re-)activate the FWF ruler whenever step 3 becomes the visible step.
  # `output$wizard_step_content` (above) rebuilds the entire step UI — and
  # with it a fresh, empty `#fwf-ruler-wrap` / SVG overlay — on every step
  # change, including backward navigation. So the JS-side ruler must be
  # told to reattach (and redraw any previously-placed boundaries) every
  # time we land on step 3, not just when advancing into it from step 2.
  enter_fwf_step3 <- function() {
    output$fwf_char_ruler_text <- renderText({ wiz$ruler_string })
    positions <- if (length(wiz$fwf_starts) > 0)
      as.list(as.integer(wiz$fwf_starts) - 1L) else list()
    session$sendCustomMessage("fwf_reinit", list(positions = positions))
  }

  observeEvent(input$wizard_back, {
    if (wiz$current_step > 1) {
      wiz$current_step <- wiz$current_step - 1L
      if (wiz$current_step == 3L) enter_fwf_step3()
    }
  })

  observeEvent(input$wizard_next, {
    sv <- step_valid()
    if (isTRUE(sv[wiz$current_step])) {
      collect_step_inputs(input, wiz, gcfg_rv())
      wiz$current_step <- wiz$current_step + 1L
      if (wiz$current_step == 3L) enter_fwf_step3()
    }
  })

  # ── Step 1 inputs ────────────────────────────────────────────────────
  observeEvent(input$wiz_dataset_name, { wiz$dataset_name <- input$wiz_dataset_name %||% "" })
  observeEvent(input$wiz_description,  { wiz$description  <- input$wiz_description  %||% "" })

  # Step 1 validation
  iv_step1 <- shinyvalidate::InputValidator$new()
  iv_step1$add_rule("wiz_dataset_name", function(v) {
    # Empty on first open: no error — Next button gate handles the required check
    if (is.null(v) || nchar(v) == 0) return(NULL)
    if (!is_valid_r_name(v))
      return("Must start with a letter; only letters, numbers, and underscores.")
    existing <- list_dataset_configs(config_dir())
    if (wiz$mode == "new" && v %in% existing)
      return(sprintf("A dataset named '%s' already exists.", v))
  })
  iv_step1$enable()

  # ── Step 2: File location ────────────────────────────────────────────
  file_roots <- c(Home=path.expand("~"), WD=getwd())

  shinyFiles::shinyDirChoose(input, "wiz_folder_browse",
    roots=file_roots, session=session)
  shinyFiles::shinyFileChoose(input, "wiz_current_file_browse",
    roots=file_roots, session=session)
  shinyFiles::shinyFileChoose(input, "wiz_prev_file_browse",
    roots=file_roots, session=session)

  # ── shinyFiles browse callbacks ─────────────────────────────────────
  # These fire when the Browse dialog is used successfully.
  # Direct text input is the primary path entry (see observers below).
  observeEvent(input$wiz_folder_browse, {
    req(is.list(input$wiz_folder_browse))
    p <- shinyFiles::parseDirPath(file_roots, input$wiz_folder_browse)
    if (length(p) > 0) {
      path <- as.character(p[1])
      updateTextInput(session, "wiz_folder_display", value=path)
      activate_folder(path)
    }
  })

  observeEvent(input$wiz_current_file_browse, {
    req(is.list(input$wiz_current_file_browse))
    p <- shinyFiles::parseFilePaths(file_roots, input$wiz_current_file_browse)
    if (nrow(p) > 0) {
      path <- as.character(p$datapath[1])
      updateTextInput(session, "wiz_current_file_display", value=path)
      activate_file(path)
    }
  })

  observeEvent(input$wiz_prev_file_browse, {
    req(is.list(input$wiz_prev_file_browse))
    p <- shinyFiles::parseFilePaths(file_roots, input$wiz_prev_file_browse)
    if (nrow(p) > 0) {
      path <- as.character(p$datapath[1])
      updateTextInput(session, "wiz_prev_file_display", value=path)
      wiz$previous_file <- path
    }
  })

  # ── Helpers: activate a folder/file path from any source ────────────
  activate_folder <- function(path) {
    path <- trimws(path)
    if (!safe_dir_exists(path)) return()
    wiz$folder <- path
    files <- list.files(path, full.names=TRUE)
    if (length(files) > 0) {
      preview <- files[order(file.mtime(files), decreasing=TRUE)[1]]
      wiz$current_preview_path <- preview
      load_raw_preview(session, preview, wiz)
    }
  }

  activate_file <- function(path) {
    path <- trimws(path)
    if (!safe_file_exists(path)) return()
    wiz$current_file        <- path
    wiz$current_preview_path <- path
    load_raw_preview(session, path, wiz)
  }

  # ── Direct text input observers (primary path entry method) ─────────
  # Debounce the folder text input so we only react after the user pauses typing.
  folder_input_r <- reactive({ input$wiz_folder_display })
  folder_input_d <- shiny::debounce(folder_input_r, 600)

  observe({
    path <- folder_input_d() %||% ""
    if (nchar(trimws(path)) > 0) activate_folder(path)
  })

  observeEvent(input$wiz_current_file_display, {
    path <- trimws(input$wiz_current_file_display %||% "")
    if (nchar(path) > 0 && safe_file_exists(path)) activate_file(path)
    else if (nchar(path) > 0) wiz$current_file <- path  # store even if not yet valid
  }, ignoreInit=TRUE)

  observeEvent(input$wiz_prev_file_display, {
    path <- trimws(input$wiz_prev_file_display %||% "")
    wiz$previous_file <- path
  }, ignoreInit=TRUE)

  # Sync file_mode radio to wiz state so build_config_list uses the correct branch
  observeEvent(input$wiz_file_mode, {
    wiz$file_mode <- input$wiz_file_mode %||% "folder"
  }, ignoreInit=TRUE)

  # ── Render Step 2 file inputs ────────────────────────────────────────
  output$wiz_file_inputs <- renderUI({
    mode <- input$wiz_file_mode %||% "folder"
    if (mode == "folder") {
      tagList(
        p(class="text-muted mb-1", style="font-size:12px;",
          "Type or paste the full folder path, or use Browse."),
        div(class="d-flex gap-2 align-items-end",
          div(style="flex:1;",
            textInput("wiz_folder_display", "Folder path",
                      value=wiz$folder %||% "", width="100%",
                      placeholder="/path/to/data/folder/")
          ),
          shinyFiles::shinyDirButton("wiz_folder_browse", "Browse",
            "Select data folder", style="height:38px;white-space:nowrap;")
        ),
        uiOutput("wiz_folder_path_status"),
        uiOutput("wiz_folder_preview_info")
      )
    } else {
      tagList(
        p(class="text-muted mb-1", style="font-size:12px;",
          "Type or paste the full file path, or use Browse."),
        div(class="d-flex gap-2 align-items-end",
          div(style="flex:1;",
            textInput("wiz_current_file_display", "Current file *",
                      value=wiz$current_file %||% "", width="100%",
                      placeholder="/path/to/data/delivery_2026.csv")
          ),
          shinyFiles::shinyFilesButton("wiz_current_file_browse", "Browse",
            "Select current file", FALSE, style="height:38px;white-space:nowrap;")
        ),
        uiOutput("wiz_current_file_status"),
        div(class="d-flex gap-2 align-items-end mt-3",
          div(style="flex:1;",
            textInput("wiz_prev_file_display", "Previous file (optional)",
                      value=wiz$previous_file %||% "", width="100%",
                      placeholder="/path/to/data/delivery_2026_prev.csv")
          ),
          shinyFiles::shinyFilesButton("wiz_prev_file_browse", "Browse",
            "Select previous file", FALSE, style="height:38px;white-space:nowrap;")
        ),
        uiOutput("wiz_prev_file_status")
      )
    }
  })

  # ── Path status badges ───────────────────────────────────────────────
  output$wiz_folder_path_status <- renderUI({
    path <- trimws(input$wiz_folder_display %||% "")
    if (nchar(path) == 0) return(NULL)
    if (safe_dir_exists(path)) {
      n <- length(list.files(path))
      div(class="text-success mt-1", style="font-size:12px;",
          sprintf("✓ Folder found  (%d files)", n))
    } else {
      div(class="text-danger mt-1", style="font-size:12px;",
          "✗ Folder not found — check the path")
    }
  })

  output$wiz_current_file_status <- renderUI({
    path <- trimws(input$wiz_current_file_display %||% "")
    if (nchar(path) == 0) return(NULL)
    if (safe_file_exists(path)) {
      sz <- format(file.size(path), big.mark=",")
      div(class="text-success mt-1", style="font-size:12px;",
          sprintf("✓ File found  (%s bytes)", sz))
    } else {
      div(class="text-danger mt-1", style="font-size:12px;",
          "✗ File not found — check the path")
    }
  })

  output$wiz_prev_file_status <- renderUI({
    path <- trimws(input$wiz_prev_file_display %||% "")
    if (nchar(path) == 0) return(NULL)
    if (safe_file_exists(path)) {
      div(class="text-success mt-1", style="font-size:12px;", "✓ File found")
    } else {
      div(class="text-warning mt-1", style="font-size:12px;",
          "⚠ File not found — leave blank to skip comparison checks")
    }
  })

  output$wiz_folder_preview_info <- renderUI({
    path <- wiz$folder %||% ""
    if (nchar(path) == 0) return(NULL)
    files <- list.files(path, full.names=TRUE)
    if (length(files) == 0) return(p(class="text-warning", "Folder is empty."))
    files_sorted <- files[order(file.mtime(files), decreasing=TRUE)]
    info <- lapply(seq_len(min(2, length(files_sorted))), function(i) {
      lbl <- if (i==1) "Current" else "Previous"
      sz  <- format(file.size(files_sorted[i]), big.mark=",")
      sprintf("%s: %s (%s bytes)", lbl, basename(files_sorted[i]), sz)
    })
    p(class="text-muted", style="font-size:12px;", paste(info, collapse=" | "))
  })

  # ── Step 4: Column Classification ────────────────────────────────────
  # Reactive dependency only on wiz$col_names (set once in step 3).
  # All other wiz fields (key_columns, expected_columns, col_types_override)
  # are isolated so that ticking a checkbox doesn't rebuild the whole table
  # and scroll the user back to row 1.
  output$step4_column_table <- renderUI({
    cols <- wiz$col_names          # reactive: rebuild only when column list changes
    req(length(cols) > 0)

    isolate({                      # isolate: checkbox/select state reads don't subscribe
      types    <- wiz$col_types_inferred
      overrides <- wiz$col_types_override
      key_cols  <- wiz$key_columns
      exp_cols  <- wiz$expected_columns

      rows <- lapply(seq_along(cols), function(i) {
        col      <- cols[i]
        inf_type <- if (i <= length(types)) types[i] else "character"
        ovr_type <- overrides[[col]] %||% "auto"
        is_key   <- col %in% key_cols
        is_exp   <- length(exp_cols) == 0 || col %in% exp_cols

        fluidRow(class="mb-1 border-bottom pb-1 align-items-center",
          column(3, tags$strong(col, style="font-size:13px;")),
          column(2, span(class="badge bg-light text-dark border", inf_type)),
          column(2,
            selectInput(paste0("s4_type_", i), NULL, width="100%",
              choices=c("auto","character","numeric","date"), selected=ovr_type)
          ),
          column(2,
            checkboxInput(paste0("s4_key_", i), "Key col", value=is_key)
          ),
          column(3,
            checkboxInput(paste0("s4_exp_", i), "Expected", value=is_exp)
          )
        )
      })

      div(
        fluidRow(class="mb-2 fw-semibold text-muted",
          column(3, "Column"), column(2, "Inferred type"),
          column(2, "Override"), column(2, "Key col"), column(3, "Expected")
        ),
        div(style="max-height:500px;overflow-y:auto;", rows)
      )
    })
  })

  # Collect step 4 inputs
  observe({
    cols <- wiz$col_names
    if (length(cols) == 0) return()
    # Bail out if step 4 UI hasn't rendered yet (inputs don't exist).
    # This prevents wiping config-loaded values during edit mode step 3 auto-parse.
    if (is.null(input[[paste0("s4_type_1")]])) return()
    key_cols <- character(0)
    exp_cols <- character(0)
    overrides <- list()
    for (i in seq_along(cols)) {
      col <- cols[i]
      type_val <- input[[paste0("s4_type_", i)]] %||% "auto"
      if (type_val != "auto") overrides[[col]] <- type_val
      if (isTRUE(input[[paste0("s4_key_", i)]])) key_cols <- c(key_cols, col)
      if (isTRUE(input[[paste0("s4_exp_", i)]])) exp_cols <- c(exp_cols, col)
    }
    wiz$col_types_override <- overrides
    wiz$key_columns <- key_cols
    wiz$expected_columns <- exp_cols
  })

  observeEvent(input$step4_select_all_expected, {
    cols <- wiz$col_names
    for (i in seq_along(cols)) updateCheckboxInput(session, paste0("s4_exp_", i), value=TRUE)
    wiz$expected_columns <- cols
  })
  observeEvent(input$step4_select_none_expected, {
    cols <- wiz$col_names
    for (i in seq_along(cols)) updateCheckboxInput(session, paste0("s4_exp_", i), value=FALSE)
    wiz$expected_columns <- character(0)
  })

  # ── Step 5: Column Rules ────────────────────────────────────────────
  # Isolate wiz field reads so accordion doesn't collapse/scroll on every
  # regex test result or allowed-value addition.
  output$step5_column_rules <- renderUI({
    cols <- wiz$col_names          # reactive dependency: col list only
    if (length(cols) == 0)
      return(p(class="text-muted fst-italic", "No columns defined yet. Complete Step 3 first."))

    isolate({
    types <- if (length(wiz$col_types_inferred) > 0) wiz$col_types_inferred else rep("character", length(cols))

    panels <- lapply(seq_along(cols), function(i) {
      col  <- cols[i]
      ctype <- if (!is.null(wiz$col_types_override[[col]])) wiz$col_types_override[[col]] else types[i]
      rules <- wiz$column_rules[[col]] %||% list()

      has_rule <- any(c(!is.null(rules$allowed_values), !is.null(rules$min_value),
                        !is.null(rules$max_value), !is.null(rules$pattern)))

      standard_fields <- tagList(
        if (ctype == "character" || ctype == "unknown") {
          existing_vals <- rules$allowed_values %||% character(0)
          div(class="mt-1",
            tags$label(paste0("Allowed values (QC-09)"), class="form-label form-label-sm"),
            selectizeInput(paste0("s5_allowed_", i), NULL,
              choices  = existing_vals,
              selected = existing_vals,
              multiple = TRUE,
              options  = list(
                create      = TRUE,
                placeholder = "Type a value, press Enter or Tab to add",
                plugins     = list("remove_button")
              ),
              width = "100%"
            )
          )
        },
        if (ctype == "numeric") {
          fluidRow(
            column(4, numericInput(paste0("s5_min_", i), "Min value (QC-10)",
                   value=rules$min_value, min=NA, max=NA)),
            column(4, numericInput(paste0("s5_max_", i), "Max value (QC-10)",
                   value=rules$max_value, min=NA, max=NA))
          )
        }
      )

      adv_fields <- tagList(
        div(class="mt-2",
          tags$label("Regex pattern (QC-13)", class="form-label form-label-sm"),
          div(class="d-flex gap-2",
            textInput(paste0("s5_pattern_", i), NULL,
                      value=rules$pattern %||% "", placeholder="e.g. ^[A-Z]{2}$", width="300px"),
            actionButton(paste0("s5_test_regex_", i), "Test",
                         class="btn btn-outline-secondary btn-sm", style="height:38px;")
          ),
          uiOutput(paste0("s5_regex_result_", i))
        ),
        fluidRow(
          column(4, numericInput(paste0("s5_maxmiss_", i), "Max missing rate",
                 value=rules$max_missing_rate, min=0, max=1, step=0.01)),
          if (ctype == "numeric") {
            column(4, numericInput(paste0("s5_maxmeanshift_", i), "Max mean shift (%)",
                   value=rules$max_numeric_mean_shift_pct, min=0, max=100, step=1))
          }
        )
      )

      bslib::accordion_panel(
        title=col,
        value=paste0("col_", i),
        tagList(
          span(class="badge bg-light text-dark border me-2", style="font-size:10px;", ctype),
          if (has_rule) span(class="badge bg-success text-white", style="font-size:10px;", "rules set"),
          div(class="mt-2", standard_fields),
          bslib::accordion(
            bslib::accordion_panel("Advanced ▼",
              value=paste0("adv_", i), adv_fields
            ),
            open=FALSE
          )
        )
      )
    })

    bslib::accordion(!!!panels, open=FALSE)
    }) # end isolate
  })

  # Regex test observers
  observe({
    cols <- wiz$col_names
    lapply(seq_along(cols), function(i) {
      local({
        ii <- i
        col <- cols[ii]
        observeEvent(input[[paste0("s5_test_regex_", ii)]], {
          pattern <- input[[paste0("s5_pattern_", ii)]] %||% ""
          if (nchar(pattern) == 0) return()
          # Test against sample
          sample_vals <- if (!is.null(wiz$csv_preview_df)) {
            v <- wiz$csv_preview_df[[col]]
            v[!is.na(v) & nchar(as.character(v)) > 0]
          } else character(0)

          result <- tryCatch({
            err_msg <- tryCatch({ grepl(pattern, "test", perl=TRUE); NULL },
                                error=function(e) e$message)
            if (!is.null(err_msg)) {
              list(type="error", msg=paste("Invalid regex:", err_msg))
            } else if (length(sample_vals) == 0) {
              list(type="info", msg="No sample data to test against.")
            } else {
              matches <- grepl(pattern, sample_vals, perl=TRUE)
              n_fail <- sum(!matches)
              if (n_fail == 0) {
                list(type="success", msg=sprintf("Pattern matches all %d non-empty values.", length(sample_vals)))
              } else {
                fail_examples <- paste(head(sample_vals[!matches], 5), collapse="', '")
                list(type="warning", msg=sprintf("Pattern does not match %d value(s): '%s'", n_fail, fail_examples))
              }
            }
          }, error=function(e) list(type="error", msg=e$message))

          output[[paste0("s5_regex_result_", ii)]] <- renderUI({
            cls <- switch(result$type, success="alert-success", warning="alert-warning",
                          error="alert-danger", info="alert-info", "alert-info")
            div(class=paste("alert p-1 mt-1", cls), style="font-size:11px;", result$msg)
          })
        })
      })
    })
  })

  # Collect step 5 inputs
  observe({
    cols <- wiz$col_names
    if (length(cols) == 0) return()
    # Bail out if step 5 UI hasn't rendered yet.
    if (is.null(input[[paste0("s5_maxmiss_1")]])) return()
    types <- wiz$col_types_inferred
    col_rules <- list()
    for (i in seq_along(cols)) {
      col   <- cols[i]
      ctype <- if (!is.null(wiz$col_types_override[[col]])) wiz$col_types_override[[col]]
               else if (i <= length(types)) types[i] else "character"
      rules <- list()
      av <- input[[paste0("s5_allowed_", i)]]
      if (!is.null(av) && length(av) > 0 && any(nchar(av) > 0)) rules$allowed_values <- av
      minv <- input[[paste0("s5_min_", i)]]
      maxv <- input[[paste0("s5_max_", i)]]
      if (!is.null(minv) && !is.na(minv)) rules$min_value <- minv
      if (!is.null(maxv) && !is.na(maxv)) rules$max_value <- maxv
      pat <- input[[paste0("s5_pattern_", i)]]
      if (!is.null(pat) && nchar(pat) > 0) rules$pattern <- pat
      miss <- input[[paste0("s5_maxmiss_", i)]]
      if (!is.null(miss) && !is.na(miss)) rules$max_missing_rate <- miss
      ms   <- input[[paste0("s5_maxmeanshift_", i)]]
      if (!is.null(ms) && !is.na(ms)) rules$max_numeric_mean_shift_pct <- ms / 100
      if (length(rules) > 0) col_rules[[col]] <- rules
    }
    wiz$column_rules <- col_rules
  })

  # ── Step 7: Custom checks ────────────────────────────────────────────
  shinyFiles::shinyFileChoose(input, "wiz_custom_browse",
    roots=file_roots, session=session)

  observeEvent(input$wiz_custom_browse, {
    req(is.list(input$wiz_custom_browse))
    p <- shinyFiles::parseFilePaths(file_roots, input$wiz_custom_browse)
    if (nrow(p) > 0) {
      updateTextInput(session, "wiz_custom_file", value=as.character(p$datapath[1]))
      wiz$custom_checks_file <- as.character(p$datapath[1])
    }
  })

  observeEvent(input$wiz_custom_file, { wiz$custom_checks_file <- input$wiz_custom_file %||% "" })

  # Debounce the custom-checks path before validating it — validation below
  # parses *and sources* the file (to confirm it defines `custom_checks`),
  # which executes the user's script. Without debouncing, that source() call
  # — and any top-level side effects in the script (package loads, file/
  # network I/O, global state changes) — would re-fire on every keystroke,
  # including for every transient partial path typed along the way. Mirrors
  # the folder_input_r/folder_input_d pattern used for the folder path above.
  custom_file_r <- reactive({ input$wiz_custom_file })
  custom_file_d <- shiny::debounce(custom_file_r, 600)

  output$step7_validation_badge <- renderUI({
    path <- custom_file_d() %||% ""
    if (nchar(path) == 0) return(NULL)
    if (!safe_file_exists(path))
      return(div(class="alert alert-danger p-1 mt-1", style="font-size:12px;",
                 "✗ File not found."))
    parse_err <- tryCatch({ parse(file=path); NULL }, error=function(e) e$message)
    if (!is.null(parse_err))
      return(div(class="alert alert-danger p-1 mt-1", style="font-size:12px;",
                 paste("✗ R syntax error:", parse_err)))
    env <- new.env(parent=baseenv())
    tryCatch(source(path, local=env), error=function(e) NULL)
    fn <- tryCatch(get("custom_checks", envir=env), error=function(e) NULL)
    if (is.null(fn) || !is.function(fn))
      return(div(class="alert alert-danger p-1 mt-1", style="font-size:12px;",
                 "✗ File must define a function named 'custom_checks'."))
    sig <- paste(names(formals(fn)), collapse=", ")
    div(class="alert alert-success p-1 mt-1", style="font-size:12px;",
        sprintf("✓ Valid — custom_checks(%s)", sig))
  })

  # ── Step 8: Review and Save ──────────────────────────────────────────
  output$step8_summary <- renderUI({
    bslib::card(bslib::card_body(
      fluidRow(
        column(6, tags$dl(
          tags$dt("Dataset"), tags$dd(wiz$dataset_name),
          tags$dt("Format"),  tags$dd(toupper(wiz$format)),
          tags$dt("Location"),tags$dd(if(nchar(wiz$folder%||%"")>0) wiz$folder else wiz$current_file)
        )),
        column(6, tags$dl(
          tags$dt("Columns"),      tags$dd(sprintf("%d total, %d expected, %d key",
                                            length(wiz$col_names),
                                            length(wiz$expected_columns),
                                            length(wiz$key_columns))),
          tags$dt("Column rules"), tags$dd(sprintf("%d configured", length(wiz$column_rules))),
          tags$dt("Custom checks"),tags$dd(if(nchar(wiz$custom_checks_file%||%"")>0) basename(wiz$custom_checks_file) else "none")
        ))
      )
    ))
  })

  output$yaml_preview <- renderText({
    yaml_preview_text(wiz, wiz$extra_keys)
  })

  observeEvent(input$wizard_save, {
    nm     <- wiz$dataset_name
    cd     <- config_dir()
    path   <- file.path(cd, paste0(nm, ".yml"))
    dir.create(cd, showWarnings=FALSE, recursive=TRUE)
    tryCatch({
      write_config(wiz, wiz$extra_keys, path)
      showNotification(sprintf("Saved: %s", path), type="message", duration=4)
      rv$wizard_open <- FALSE
      rv$active_section <- "datasets"
      rv$active_dataset <- nm
      rv$dataset_refresh <- Sys.time()
    }, error=function(e) {
      showModal(modalDialog(
        title="Save failed",
        paste("Could not write config:", e$message),
        easyClose=TRUE
      ))
    })
  })

  wiz
}

# Helper to collect inputs from various steps
collect_step_inputs <- function(input, wiz, gcfg) {
  # Bail if step 6 UI hasn't rendered — prevents wiping config-loaded rule_overrides
  # when Next is clicked on steps 1-5 (same guard pattern as step 4/5 collectors).
  if (is.null(input$wiz_ro_max_missing)) return()

  # Rule overrides from step 6
  dr <- gcfg$default_rules %||% list()
  overrides <- list()
  # Returns the input value if it is set and differs from the default, else NULL.
  # Assigning NULL to an absent list key is a no-op, so only changed keys persist
  # (a pure helper — avoids the `<<-` an accumulating closure would need).
  chk_num <- function(input_id, dflt) {
    v <- input[[input_id]]
    if (!is.null(v) && !is.na(v) && !identical(v, dflt)) v else NULL
  }
  overrides[["max_missing_rate"]]               <- chk_num("wiz_ro_max_missing",    dr$max_missing_rate %||% 0.05)
  overrides[["max_non_numeric_rate"]]           <- chk_num("wiz_ro_max_nonnumeric", dr$max_non_numeric_rate %||% 0.01)
  overrides[["min_row_count"]]                  <- chk_num("wiz_ro_min_rows",       dr$min_row_count %||% 0)
  v <- input$wiz_ro_max_rowchg;   dflt <- round((dr$max_row_count_change_pct %||% 0.10)*100,2)
  if (!is.null(v) && !is.na(v) && !identical(v, dflt)) overrides$max_row_count_change_pct <- v/100
  v <- input$wiz_ro_max_meanshift; dflt <- round((dr$max_numeric_mean_shift_pct %||% 0.20)*100,2)
  if (!is.null(v) && !is.na(v) && !identical(v, dflt)) overrides$max_numeric_mean_shift_pct <- v/100
  overrides[["max_missing_rate_change_pp"]]     <- chk_num("wiz_ro_max_misschg",   dr$max_missing_rate_change_pp %||% 2.0)
  overrides[["max_non_numeric_rate_change_pp"]] <- chk_num("wiz_ro_max_nonnumchg", dr$max_non_numeric_rate_change_pp %||% 1.0)
  overrides[["type_inference_threshold"]]       <- chk_num("wiz_ro_type_inf",      dr$type_inference_threshold %||% 0.90)
  for (pair in list(
    list("flag_new_columns","wiz_ro_flag_new",TRUE),
    list("flag_dropped_columns","wiz_ro_flag_drop",TRUE),
    list("flag_type_changes","wiz_ro_flag_type",TRUE),
    list("flag_column_order_change","wiz_ro_flag_order",TRUE)
  )) {
    v <- input[[pair[[2]]]]; dflt <- pair[[3]]
    if (!is.null(v) && !identical(isTRUE(v), dflt)) overrides[[pair[[1]]]] <- isTRUE(v)
  }
  wiz$rule_overrides <- overrides
}

# Global config server logic (spec §19)

server_global <- function(input, output, session, rv, config_dir, gcfg_rv) {

  roots <- c(Home=path.expand("~"), WD=getwd())

  shinyFiles::shinyFileChoose(input, "gcfg_db_browse",
    roots=roots, session=session)
  shinyFiles::shinyDirChoose(input, "gcfg_dir_browse",
    roots=roots, session=session)

  observeEvent(input$gcfg_db_browse, {
    req(is.list(input$gcfg_db_browse))
    p <- shinyFiles::parseFilePaths(roots, input$gcfg_db_browse)
    if (nrow(p) > 0) updateTextInput(session, "gcfg_snapshot_db", value=as.character(p$datapath[1]))
  })

  observeEvent(input$gcfg_dir_browse, {
    req(is.list(input$gcfg_dir_browse))
    p <- shinyFiles::parseDirPath(roots, input$gcfg_dir_browse)
    if (length(p) > 0) updateTextInput(session, "gcfg_report_dir", value=as.character(p[1]))
  })

  # Validation
  iv_gcfg <- shinyvalidate::InputValidator$new()
  iv_gcfg$add_rule("gcfg_snapshot_db", function(v) {
    if (is.null(v) || !is.character(v) || nchar(v) == 0) return(NULL)
    parent <- dirname(v)
    if (!safe_dir_exists(parent)) sprintf("Directory not found: %s", parent)
  })
  iv_gcfg$add_rule("gcfg_report_dir", function(v) {
    if (is.null(v) || !is.character(v) || nchar(v) == 0) return(NULL)
    if (!safe_dir_exists(v)) "Directory not found. Create it first or change path."
  })
  iv_gcfg$enable()

  # Render validation message outputs (replacing sv_output which doesn't exist)
  output$sv_gcfg_snapshot_db <- renderUI({
    v <- input$gcfg_snapshot_db
    if (is.null(v) || nchar(v) == 0) return(NULL)
    parent <- dirname(v)
    if (!safe_dir_exists(parent))
      tags$div(class="text-danger", style="font-size:12px;",
               sprintf("Directory not found: %s", parent))
  })
  output$sv_gcfg_report_dir <- renderUI({
    v <- input$gcfg_report_dir
    if (is.null(v) || nchar(v) == 0) return(NULL)
    if (!safe_dir_exists(v))
      tags$div(class="text-danger", style="font-size:12px;",
               "Directory not found. Create it first or change path.")
  })

  observeEvent(input$gcfg_save, {
    if (!iv_gcfg$is_valid()) {
      showNotification("Fix validation errors before saving.", type="error")
      return()
    }
    gcfg <- list(
      snapshot_db       = input$gcfg_snapshot_db,
      report_output_dir = input$gcfg_report_dir,
      default_rules     = list(
        type_inference_threshold       = input$gcfg_type_inf_threshold,
        max_missing_rate               = input$gcfg_max_missing_rate,
        max_non_numeric_rate           = input$gcfg_max_non_numeric_rate,
        min_row_count                  = as.integer(input$gcfg_min_row_count),
        max_row_count_change_pct       = input$gcfg_max_row_count_chg / 100,
        max_numeric_mean_shift_pct     = input$gcfg_max_mean_shift / 100,
        max_missing_rate_change_pp     = input$gcfg_max_missing_chg,
        max_non_numeric_rate_change_pp = input$gcfg_max_nonnumeric_chg,
        flag_new_columns               = isTRUE(input$gcfg_flag_new_cols),
        flag_dropped_columns           = isTRUE(input$gcfg_flag_drop_cols),
        flag_type_changes              = isTRUE(input$gcfg_flag_type_chg),
        flag_column_order_change       = isTRUE(input$gcfg_flag_col_order)
      )
    )
    tryCatch({
      write_global_config(gcfg, global_config_path(config_dir()))
      register_report_resource_path(gcfg$report_output_dir, config_dir())
      gcfg_rv(gcfg)
      showNotification("Global config saved.", type="message", duration=3)
    }, error = function(e) {
      showModal(modalDialog(
        title="Save failed",
        paste("Could not write global config:", e$message),
        easyClose=TRUE
      ))
    })
  })
}

# Shared utilities

`%||%` <- function(a, b) if (!is.null(a) && length(a) > 0 && !is.na(a[1]) && a[1] != "") a else b

make_report_filename <- function(dataset_name, run_timestamp) {
  ts_raw  <- gsub("[^0-9]", "", substr(run_timestamp, 1, 19))
  ts_slug <- paste0(substr(ts_raw, 1, 8), "_", substr(ts_raw, 9, 14))
  paste0(dataset_name, "_", ts_slug, ".html")
}

# Escape a string for safe embedding inside a single-quoted JavaScript string
# literal — e.g. the dataset name interpolated into
# onclick="Shiny.setInputValue('ds_action', {ds:'<name>', ...})". Backslashes
# are escaped first so the backslash introduced for an escaped quote isn't
# itself re-escaped by the second gsub. `is_valid_r_name()` keeps dataset
# names alnum/underscore-only when created via the wizard, but names are
# read back from filenames on disk (list_dataset_configs()) without
# re-validation, so a manually placed/renamed config could carry characters
# that would otherwise break out of the JS string literal.
js_string_escape <- function(x) {
  x <- gsub("\\\\", "\\\\\\\\", x)
  gsub("'", "\\\\'", x)
}

.status_cfg <- function(status) {
  switch(as.character(status),
    PASS    = list(bg="#5cb85c", sym="✓", text="PASS"),
    WARN    = list(bg="#f0ad4e", sym="⚠", text="WARN"),
    FAIL    = list(bg="#d9534f", sym="✗", text="FAIL"),
    RUNNING = list(bg="#337ab7", sym="●", text="RUNNING"),
    list(bg="#999999", sym="—", text=as.character(status %||% ""))
  )
}

status_badge <- function(status) {
  cfg <- .status_cfg(status)
  tags$span(
    style=sprintf("background:%s;color:#fff;padding:2px 8px;border-radius:4px;font-size:12px;font-weight:600;white-space:nowrap;", cfg$bg),
    paste(cfg$sym, cfg$text)
  )
}

status_badge_html <- function(status) {
  cfg <- .status_cfg(status)
  sprintf('<span style="background:%s;color:#fff;padding:2px 8px;border-radius:4px;font-size:12px;font-weight:600;">%s %s</span>',
          cfg$bg, cfg$sym, cfg$text)
}

safe_dir_exists <- function(path) {
  tryCatch(isTRUE(dir.exists(path)), error = function(e) FALSE)
}

safe_file_exists <- function(path) {
  tryCatch(isTRUE(file.exists(path)), error = function(e) FALSE)
}

is_valid_r_name <- function(x) {
  grepl("^[a-zA-Z][a-zA-Z0-9_]*$", x)
}

# Coerce an arbitrary header token into a syntactically valid R name:
# non-word chars → "_", collapse/trim underscores, prefix "col_" if it does
# not start with a letter. Always returns something passing is_valid_r_name().
sanitize_r_name <- function(x) {
  x <- trimws(x %||% "")
  x <- gsub("[^A-Za-z0-9_]", "_", x)
  x <- gsub("_+", "_", x)
  x <- sub("^_+", "", x)
  x <- sub("_+$", "", x)
  if (x == "" || !grepl("^[A-Za-z]", x)) x <- paste0("col_", x)
  x <- sub("_+$", "", x)
  if (x == "" || x == "col_") x <- "col"
  x
}

# Given the raw (unmangled) header names of a CSV, produce a parallel list of
# valid, unique suggestions plus a human-readable reason for each change.
# Duplicate names get a positional suffix in first-appearance order
# (1st → base, 2nd → base_2, 3rd → base_3); names invalid for other reasons
# are sanitised. A final pass guarantees no suggestion collides with another.
suggest_col_names <- function(raw_names) {
  raw_names <- as.character(raw_names)
  n <- length(raw_names)
  out    <- character(n)
  reason <- character(n)
  taken  <- character(0)
  for (i in seq_len(n)) {
    raw  <- raw_names[i]
    base <- sanitize_r_name(raw)
    why  <- if (!identical(base, raw)) "invalid name — sanitised" else ""
    occ  <- sum(raw_names[seq_len(i)] == raw)   # 1 for first, 2 for second, ...
    cand <- base
    if (occ > 1L) {
      cand <- paste0(base, "_", occ)
      why  <- sprintf("duplicate of column %d — suggested %s",
                      match(raw, raw_names), cand)
    }
    bump <- occ
    while (cand %in% taken) {
      bump <- bump + 1L
      cand <- paste0(base, "_", bump)
      if (why == "") why <- sprintf("name collision — suggested %s", cand)
    }
    out[i]    <- cand
    reason[i] <- why
    taken     <- c(taken, cand)
  }
  list(names = out, reason = reason)
}

# Does the CSV at step 3 need the column-naming editor shown?
# - headerless files: always (names are placeholders the user should set)
# - header files: only when a raw header name is invalid or duplicated
csv_needs_naming <- function(wiz, has_header) {
  if (!isTRUE(has_header)) return(length(wiz$csv_col_names_detected) > 0)
  raw <- wiz$raw_header_names
  if (length(raw) == 0) return(FALSE)
  any(!is_valid_r_name(raw)) || any(duplicated(raw))
}

list_dataset_configs <- function(config_dir) {
  if (!safe_dir_exists(config_dir)) return(character(0))
  files <- list.files(config_dir, pattern="\\.yml$", full.names=FALSE)
  files <- files[files != "dqcheckr.yml"]
  sort(tools::file_path_sans_ext(files))
}

read_snapshot_history <- function(db_path, dataset_name = NULL, n = 10) {
  empty <- data.frame(
    id=integer(0), dataset_name=character(0), file_name=character(0),
    run_timestamp=character(0), overall_status=character(0),
    check_fail_count=integer(0), check_warn_count=integer(0),
    row_count=integer(0), stringsAsFactors=FALSE
  )
  if (is.null(db_path) || db_path == "" || !safe_file_exists(db_path)) return(empty)

  tryCatch({
    con <- DBI::dbConnect(RSQLite::SQLite(), db_path)
    on.exit(DBI::dbDisconnect(con), add=TRUE)

    if (!is.null(dataset_name) && dataset_name != "") {
      DBI::dbGetQuery(con,
        "SELECT id, dataset_name, file_name, run_timestamp,
                overall_status, check_fail_count, check_warn_count, row_count
         FROM snapshots WHERE dataset_name = ? ORDER BY id DESC LIMIT ?",
        list(dataset_name, as.integer(n)))
    } else {
      DBI::dbGetQuery(con,
        "SELECT id, dataset_name, file_name, run_timestamp,
                overall_status, check_fail_count, check_warn_count, row_count
         FROM snapshots ORDER BY id DESC LIMIT ?",
        list(as.integer(n)))
    }
  }, error = function(e) {
    message("read_snapshot_history: query failed for db_path '", db_path,
            "': ", conditionMessage(e))
    empty
  })
}

read_all_snapshot_history <- function(db_path, n = 200) {
  read_snapshot_history(db_path, dataset_name=NULL, n=n)
}

# Infer column types from a character vector sample (dqcheckr logic)
infer_col_type_simple <- function(x) {
  x <- x[!is.na(x) & x != ""]
  if (length(x) == 0) return("unknown")
  date_fmts <- c("%Y-%m-%d","%d/%m/%Y","%m/%d/%Y","%Y%m%d","%d-%m-%Y")
  for (fmt in date_fmts) {
    parsed <- suppressWarnings(as.Date(x, format=fmt))
    if (all(!is.na(parsed))) return("date")
  }
  numeric_ok <- suppressWarnings(!is.na(as.numeric(x)))
  if (mean(numeric_ok) >= 0.90) return("numeric")
  "character"
}

# Generate ruler string for N characters
make_ruler_string <- function(max_chars = 120) {
  ruler <- rep(" ", max_chars)
  for (i in seq(5, max_chars, by=5)) {
    label <- as.character(i)
    pos <- i
    for (j in seq_along(strsplit(label,"")[[1]])) {
      idx <- pos - nchar(label) + j
      if (idx >= 1 && idx <= max_chars) ruler[idx] <- substr(label, j, j)
    }
    if (pos <= max_chars) ruler[pos] <- "|"
  }
  paste(c("|", ruler), collapse="")
}

global_config_path <- function(config_dir) {
  file.path(config_dir, "dqcheckr.yml")
}

# TRUE if `p` is an absolute path: unix "/" or "~", a Windows drive ("C:\" /
# "C:/"), or a UNC share ("\\server").
is_absolute_path <- function(p) {
  grepl("^(/|~|[A-Za-z]:[\\\\/]|\\\\\\\\)", p %||% "")
}

# Deployment root — the directory that relative infra paths (`snapshot_db`,
# `report_output_dir`) are based on. dqcheckr's CLI resolves those relative to
# the directory it is run from, which by convention is the deployment root that
# contains config/, data/ and reports/; `config_dir` is `<root>/config`. The GUI
# must NOT use getwd() for this, because shiny::runApp() changes the working
# directory to the installed app folder — so we anchor to the parent of the
# config directory instead. (See also the matching `wd =` on the callr runs.)
deployment_root <- function(config_dir) {
  dirname(config_dir %||% ".")
}

# Resolve a (possibly relative) infra path against the deployment root. Absolute
# paths are returned unchanged, so an absolute snapshot_db/report_output_dir in
# the config also works.
resolve_infra_path <- function(path, config_dir, default = NULL, mustWork = FALSE) {
  path <- path %||% default %||% ""
  if (nchar(path) == 0) return("")
  if (!is_absolute_path(path)) path <- file.path(deployment_root(config_dir), path)
  normalizePath(path, mustWork = mustWork)
}

# Register (or re-register) the reports directory as the "dq_reports" static
# resource path. Re-registering with a new path replaces the old mapping, so
# this must be called again whenever report_output_dir changes at runtime.
register_report_resource_path <- function(report_output_dir, config_dir) {
  report_dir <- resolve_infra_path(report_output_dir, config_dir,
                                   default = "reports/", mustWork = FALSE)
  if (nchar(report_dir) > 0 && dir.exists(report_dir)) {
    addResourcePath("dq_reports", report_dir)
  }
}

# Convert a UTC ISO timestamp (from the snapshot DB) to a local-time display string.
utc_to_local_display <- function(ts) {
  parsed <- as.POSIXct(ts, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
  format(parsed, format = "%Y-%m-%d %H:%M:%S", tz = "")
}

# Unit tests for R/utils.R


# ── %||% ─────────────────────────────────────────────────────────────────────

test_that("%||% returns left when it has a real value", {
  expect_equal("hello"   %||% "default", "hello")
  expect_equal(42L       %||% 0L,        42L)
  expect_equal(FALSE     %||% TRUE,      FALSE)
})

test_that("%||% returns right when left is NULL", {
  expect_equal(NULL %||% "default", "default")
  expect_equal(NULL %||% 0,         0)
})

test_that("%||% returns right when left is NA", {
  expect_equal(NA      %||% "default", "default")
  expect_equal(NA_real_ %||% 99,       99)
})

test_that("%||% returns right when left is empty string", {
  expect_equal("" %||% "fallback", "fallback")
})

test_that("%||% returns right when left is zero-length vector", {
  expect_equal(character(0) %||% "x", "x")
})

# ── is_valid_r_name ───────────────────────────────────────────────────────────

test_that("is_valid_r_name accepts valid single-word names", {
  expect_true(is_valid_r_name("dataset"))
  expect_true(is_valid_r_name("MyDataset"))
  expect_true(is_valid_r_name("dataset_1"))
  expect_true(is_valid_r_name("x"))
  expect_true(is_valid_r_name("CamelCase123"))
  expect_true(is_valid_r_name("with_under_score"))
})

test_that("is_valid_r_name rejects names starting with a digit", {
  expect_false(is_valid_r_name("1dataset"))
  expect_false(is_valid_r_name("123"))
  expect_false(is_valid_r_name("0x"))
})

test_that("is_valid_r_name rejects names with hyphens", {
  expect_false(is_valid_r_name("my-dataset"))
  expect_false(is_valid_r_name("a-b"))
})

test_that("is_valid_r_name rejects names with spaces", {
  expect_false(is_valid_r_name("my dataset"))
  expect_false(is_valid_r_name(" leading"))
})

test_that("is_valid_r_name rejects names with dots (UI hint says letters/numbers/underscores only)", {
  expect_false(is_valid_r_name("my.dataset"))
})

test_that("is_valid_r_name rejects empty string", {
  expect_false(is_valid_r_name(""))
})

# ── list_dataset_configs ──────────────────────────────────────────────────────

test_that("list_dataset_configs returns names in alphabetical order", {
  dir <- withr::local_tempdir()
  # Create files in non-alphabetical order
  file.create(file.path(dir, "zebra.yml"))
  file.create(file.path(dir, "alpha.yml"))
  file.create(file.path(dir, "mango.yml"))
  file.create(file.path(dir, "beta.yml"))

  result <- list_dataset_configs(dir)
  expect_equal(result, c("alpha", "beta", "mango", "zebra"))
})

test_that("list_dataset_configs excludes dqcheckr.yml", {
  dir <- withr::local_tempdir()
  file.create(file.path(dir, "my_ds.yml"))
  file.create(file.path(dir, "dqcheckr.yml"))

  result <- list_dataset_configs(dir)
  expect_equal(result, "my_ds")
  expect_false("dqcheckr" %in% result)
})

test_that("list_dataset_configs returns character(0) for empty directory", {
  dir <- withr::local_tempdir()
  expect_equal(list_dataset_configs(dir), character(0))
})

test_that("list_dataset_configs returns character(0) for non-existent path", {
  expect_equal(list_dataset_configs("/this/path/does/not/exist/xyz123"), character(0))
})

test_that("list_dataset_configs ignores non-yml files", {
  dir <- withr::local_tempdir()
  file.create(file.path(dir, "dataset.yml"))
  file.create(file.path(dir, "notes.txt"))
  file.create(file.path(dir, "data.csv"))

  result <- list_dataset_configs(dir)
  expect_equal(result, "dataset")
})

test_that("list_dataset_configs result is consistently sorted regardless of filesystem order", {
  dir <- withr::local_tempdir()
  # Create in reverse alphabetical order
  for (nm in rev(c("aardvark", "elephant", "mongoose", "zebra"))) {
    file.create(file.path(dir, paste0(nm, ".yml")))
    Sys.sleep(0.01)  # ensure different mtime
  }
  result <- list_dataset_configs(dir)
  expect_equal(result, c("aardvark", "elephant", "mongoose", "zebra"))
})

# ── infer_col_type_simple ─────────────────────────────────────────────────────

test_that("infer_col_type_simple detects numeric columns", {
  expect_equal(infer_col_type_simple(c("1", "2.5", "100", "-3.14")), "numeric")
})

test_that("infer_col_type_simple detects numeric with sparse NAs", {
  # 90% threshold — 9 out of 10 valid
  x <- c("1","2","3","4","5","6","7","8","9","not_a_number")
  expect_equal(infer_col_type_simple(x), "numeric")
})

test_that("infer_col_type_simple detects ISO date columns", {
  expect_equal(infer_col_type_simple(c("2024-01-01","2024-06-15","2023-12-31")), "date")
})

test_that("infer_col_type_simple detects d/m/Y date format", {
  expect_equal(infer_col_type_simple(c("01/01/2024","15/06/2024","31/12/2023")), "date")
})

test_that("infer_col_type_simple returns character for text columns", {
  expect_equal(infer_col_type_simple(c("Alice","Bob","Carol","Dave")), "character")
})

test_that("infer_col_type_simple returns unknown for all-NA input", {
  expect_equal(infer_col_type_simple(c(NA, NA, NA)), "unknown")
})

test_that("infer_col_type_simple returns unknown for zero-length input", {
  expect_equal(infer_col_type_simple(character(0)), "unknown")
})

test_that("infer_col_type_simple ignores NA and empty string when inferring", {
  # Should still detect numeric if non-empty values are numeric
  expect_equal(infer_col_type_simple(c("1", "2", NA, "", "3")), "numeric")
})

# ── safe_file_exists / safe_dir_exists ───────────────────────────────────────

test_that("safe_file_exists returns TRUE for existing file", {
  f <- tempfile()
  file.create(f)
  on.exit(unlink(f))
  expect_true(safe_file_exists(f))
})

test_that("safe_file_exists returns FALSE for missing file", {
  expect_false(safe_file_exists("/this/does/not/exist/xyz.csv"))
})

test_that("safe_file_exists returns FALSE for NULL input", {
  expect_false(safe_file_exists(NULL))
})

test_that("safe_dir_exists returns TRUE for existing directory", {
  expect_true(safe_dir_exists(tempdir()))
})

test_that("safe_dir_exists returns FALSE for missing directory", {
  expect_false(safe_dir_exists("/no/such/directory/xyz"))
})

# ── make_ruler_string ─────────────────────────────────────────────────────────

test_that("make_ruler_string starts with a pipe character", {
  expect_true(startsWith(make_ruler_string(80), "|"))
})

test_that("make_ruler_string has length max_chars + 1 (leading pipe)", {
  expect_equal(nchar(make_ruler_string(80)),  81L)
  expect_equal(nchar(make_ruler_string(120)), 121L)
  expect_equal(nchar(make_ruler_string(40)),  41L)
})

test_that("make_ruler_string has pipe markers at every 5th character position", {
  # The ruler writes digits left-aligned before each pipe; the last digit of
  # multi-digit labels is overwritten by "|" (e.g. "1|" at position 10, not "10").
  # Test structural properties rather than literal substrings.
  ruler <- make_ruler_string(60)
  chars <- strsplit(ruler, "")[[1]]

  # Leading pipe at index 1
  expect_equal(chars[1], "|")

  # Pipe markers fall at string indices 6, 11, 16, 21 ... (char positions 5, 10, 15, 20 ...)
  pipe_idx <- which(chars == "|")
  expect_true(6  %in% pipe_idx, info = "pipe expected at char position 5")
  expect_true(11 %in% pipe_idx, info = "pipe expected at char position 10")
  expect_true(21 %in% pipe_idx, info = "pipe expected at char position 20")
  expect_true(51 %in% pipe_idx, info = "pipe expected at char position 50")

  # At least one numeric digit appears in the ruler
  expect_true(any(chars %in% as.character(1:9)))
})

# ── status_badge_html ─────────────────────────────────────────────────────────

test_that("status_badge_html returns HTML string for known statuses", {
  expect_match(status_badge_html("PASS"), "PASS")
  expect_match(status_badge_html("FAIL"), "FAIL")
  expect_match(status_badge_html("WARN"), "WARN")
})

test_that("status_badge_html returns HTML span for unknown status", {
  result <- status_badge_html("UNKNOWN_STATUS")
  expect_match(result, "<span")
})

# ── utc_to_local_display ──────────────────────────────────────────────────────
#
# Production (dqr/dqcheckr/R/snapshot.R:211) writes run_timestamp to the
# snapshot DB as format(Sys.time(), "%Y-%m-%dT%H:%M:%SZ", tz = "UTC"), e.g.
# "2026-05-31T12:08:08Z". utc_to_local_display() parses *that* format with
# as.POSIXct(..., tz = "UTC") and re-renders it in the session's local
# timezone. Because the result depends on Sys.timezone(), each test fixes
# the timezone with withr::local_timezone() so the expected string is
# deterministic regardless of where the suite runs (and so DST edge cases
# can be exercised deliberately).

test_that("utc_to_local_display converts a production-format timestamp to local time (no DST)", {
  withr::local_timezone("Australia/Sydney")
  # 2026-06-01 falls in Sydney's non-DST period (AEST, UTC+10)
  expect_equal(utc_to_local_display("2026-06-01T00:00:00Z"), "2026-06-01 10:00:00")
})

test_that("utc_to_local_display converts a production-format timestamp to local time (DST)", {
  withr::local_timezone("America/New_York")
  # 2026-06-01 falls in New York's DST period (EDT, UTC-4)
  expect_equal(utc_to_local_display("2026-06-01T12:00:00Z"), "2026-06-01 08:00:00")
})

test_that("utc_to_local_display vectorises over multiple timestamps", {
  withr::local_timezone("UTC")
  result <- utc_to_local_display(c("2026-05-31T12:08:08Z", "2026-06-01T00:00:00Z"))
  expect_equal(result, c("2026-05-31 12:08:08", "2026-06-01 00:00:00"))
})

test_that("utc_to_local_display returns NA for missing or unparseable input", {
  withr::local_timezone("UTC")
  expect_true(is.na(utc_to_local_display(NA_character_)))
  expect_true(is.na(utc_to_local_display("not-a-timestamp")))
})

test_that("utc_to_local_display does not parse the space-separated timestamps used by some test fixtures", {
  # Several fixtures elsewhere in this suite (e.g. test-history.R) use
  # space-separated timestamps like "2026-05-31 12:08:08" rather than the
  # production "T...Z" form. utc_to_local_display only understands the
  # production format — a space-separated string silently parses to NA
  # rather than producing a converted display string. Pinning that behaviour
  # here so a future fixture rewrite that switches to production-format
  # timestamps doesn't go unnoticed (and so nobody "fixes" this function to
  # also accept the space form without realising real data never uses it).
  withr::local_timezone("UTC")
  expect_true(is.na(utc_to_local_display("2026-05-31 12:08:08")))
})

# ── sanitize_r_name / suggest_col_names / csv_needs_naming (B-08) ─────────────

test_that("sanitize_r_name always yields a valid R name", {
  expect_true(is_valid_r_name(sanitize_r_name("Payee Name")))
  expect_equal(sanitize_r_name("Payee Name"), "Payee_Name")
  expect_equal(sanitize_r_name("123abc"),     "col_123abc")
  expect_equal(sanitize_r_name("$$$"),        "col")
  expect_equal(sanitize_r_name(" a-b/c "),    "a_b_c")
  expect_true(is_valid_r_name(sanitize_r_name("")))
})

test_that("suggest_col_names de-duplicates with positional suffixes", {
  s <- suggest_col_names(c("PayeeName", "Amount", "PayeeName", "PayeeName"))
  expect_equal(s$names, c("PayeeName", "Amount", "PayeeName_2", "PayeeName_3"))
  expect_equal(s$reason[1], "")               # first occurrence unchanged
  expect_match(s$reason[3], "duplicate of column 1")
  # every suggestion is valid and unique — the whole point
  expect_true(all(is_valid_r_name(s$names)))
  expect_false(any(duplicated(s$names)))
})

test_that("suggest_col_names sanitises invalid-but-unique names", {
  s <- suggest_col_names(c("Payee Name", "1st", "ok_name"))
  expect_equal(s$names, c("Payee_Name", "col_1st", "ok_name"))
  expect_match(s$reason[1], "sanitised")
  expect_equal(s$reason[3], "")
})

test_that("suggest_col_names breaks collisions created by suffixing", {
  # "a" duplicated → second becomes "a_2"; a literal "a_2" already present must
  # not collide.
  s <- suggest_col_names(c("a", "a", "a_2"))
  expect_false(any(duplicated(s$names)))
  expect_true(all(is_valid_r_name(s$names)))
})

test_that("csv_needs_naming triggers on duplicate/invalid headers only", {
  ok  <- list(csv_col_names_detected = c("id", "name"),
              raw_header_names       = c("id", "name"))
  dup <- list(csv_col_names_detected = c("id", "id"),
              raw_header_names       = c("id", "id"))
  bad <- list(csv_col_names_detected = c("id", "a b"),
              raw_header_names       = c("id", "a b"))
  expect_false(csv_needs_naming(ok,  has_header = TRUE))
  expect_true( csv_needs_naming(dup, has_header = TRUE))
  expect_true( csv_needs_naming(bad, has_header = TRUE))
  # headerless always needs naming when columns are detected
  expect_true( csv_needs_naming(ok,  has_header = FALSE))
})

# ── infra path resolution (deployment-root anchoring; the GUI history bug) ────

test_that("is_absolute_path recognises absolute vs relative paths", {
  expect_true(is_absolute_path("/var/data/snapshots.sqlite"))
  expect_true(is_absolute_path("~/data"))
  expect_false(is_absolute_path("data/snapshots.sqlite"))
  expect_false(is_absolute_path("reports/"))
  expect_false(is_absolute_path(""))
})

test_that("deployment_root is the parent of the config dir", {
  expect_equal(deployment_root("/srv/deploy/config"), "/srv/deploy")
})

test_that("resolve_infra_path anchors relative paths to the deployment root (not getwd)", {
  # This is the core of the GUI history bug: shiny::runApp() changes getwd() to
  # the installed app dir, so a relative snapshot_db must be anchored to the
  # deployment root (parent of config_dir), not the working directory.
  root    <- withr::local_tempdir()
  dir.create(file.path(root, "config"))
  dir.create(file.path(root, "data"))
  cfg_dir <- file.path(root, "config")
  got     <- resolve_infra_path("data/snapshots.sqlite", cfg_dir,
                                default = "data/snapshots.sqlite")
  expect_equal(got, normalizePath(file.path(root, "data", "snapshots.sqlite"),
                                  mustWork = FALSE))
})

test_that("resolve_infra_path leaves absolute paths unchanged", {
  abs <- withr::local_tempfile(fileext = ".sqlite")
  file.create(abs)
  expect_equal(resolve_infra_path(abs, "/some/config"),
               normalizePath(abs, mustWork = FALSE))
})

test_that("resolve_infra_path falls back to default when path is empty/NULL", {
  root    <- withr::local_tempdir()
  dir.create(file.path(root, "config"))
  cfg_dir <- file.path(root, "config")
  expect_equal(resolve_infra_path(NULL, cfg_dir, default = "data/snapshots.sqlite"),
               normalizePath(file.path(root, "data", "snapshots.sqlite"),
                             mustWork = FALSE))
})

# Unit tests for history utilities and report URL construction

library(shiny)
library(DBI)
library(RSQLite)

# ── Test DB helper ────────────────────────────────────────────────────────────

make_test_snapshot_db <- function(envir = parent.frame()) {
  tmp <- tempfile(fileext = ".sqlite")
  withr::defer(unlink(tmp), envir = envir)
  con <- DBI::dbConnect(RSQLite::SQLite(), tmp)
  DBI::dbExecute(con, "
    CREATE TABLE snapshots (
      id INTEGER PRIMARY KEY,
      dataset_name TEXT,
      file_name TEXT,
      run_timestamp TEXT,
      overall_status TEXT,
      check_fail_count INTEGER,
      check_warn_count INTEGER,
      row_count INTEGER
    )")
  DBI::dbExecute(con, "
    INSERT INTO snapshots VALUES
      (1, 'ds_a', 'file_a1.csv', '2026-05-31 10:00:00', 'PASS', 0, 0, 100),
      (2, 'ds_a', 'file_a2.csv', '2026-05-31 11:00:00', 'FAIL', 2, 1,  98),
      (3, 'ds_b', 'file_b1.csv', '2026-05-31 12:00:00', 'WARN', 0, 1, 200),
      (4, 'ds_b', 'file_b2.csv', '2026-05-31 13:00:00', 'PASS', 0, 0, 202)")
  DBI::dbDisconnect(con)
  tmp
}

# ── read_snapshot_history: bad paths ─────────────────────────────────────────

test_that("read_snapshot_history returns empty data frame for missing db path", {
  result <- read_snapshot_history("/no/such/path.sqlite", n = 10)
  expect_equal(nrow(result), 0L)
  expect_true("id" %in% names(result))
  expect_true("dataset_name" %in% names(result))
})

test_that("read_snapshot_history returns empty data frame for empty string path", {
  result <- read_snapshot_history("", n = 10)
  expect_equal(nrow(result), 0L)
})

test_that("read_snapshot_history returns empty data frame for NULL path", {
  result <- read_snapshot_history(NULL, n = 10)
  expect_equal(nrow(result), 0L)
})

# ── read_snapshot_history: correct data ───────────────────────────────────────

test_that("read_snapshot_history returns all rows when dataset_name is NULL", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, dataset_name = NULL, n = 100)
  expect_equal(nrow(result), 4L)
})

test_that("read_snapshot_history filters correctly by dataset_name", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, dataset_name = "ds_a", n = 100)
  expect_equal(nrow(result), 2L)
  expect_true(all(result$dataset_name == "ds_a"))
})

test_that("read_snapshot_history returns empty frame for unknown dataset_name", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, dataset_name = "no_such_ds", n = 100)
  expect_equal(nrow(result), 0L)
})

test_that("read_snapshot_history respects n limit", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, dataset_name = NULL, n = 2)
  expect_equal(nrow(result), 2L)
})

test_that("read_snapshot_history returns rows in descending id order", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, dataset_name = NULL, n = 100)
  expect_equal(result$id, c(4L, 3L, 2L, 1L))
})

test_that("read_snapshot_history n limit returns the most recent rows (highest ids)", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, dataset_name = NULL, n = 2)
  # With DESC + LIMIT 2 we expect the two highest ids
  expect_equal(sort(result$id), c(3L, 4L))
})

test_that("read_snapshot_history returns correct column names", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, n = 1)
  expected_cols <- c("id", "dataset_name", "file_name", "run_timestamp",
                     "overall_status", "check_fail_count", "check_warn_count", "row_count")
  expect_true(all(expected_cols %in% names(result)))
})

test_that("read_snapshot_history returns correct field values", {
  db <- make_test_snapshot_db()
  result <- read_snapshot_history(db, dataset_name = "ds_a", n = 100)
  # The most recent ds_a row (id=2) should be first
  expect_equal(result$id[1], 2L)
  expect_equal(result$overall_status[1], "FAIL")
  expect_equal(result$check_fail_count[1], 2L)
  expect_equal(result$row_count[1], 98L)
})

# ── read_all_snapshot_history ─────────────────────────────────────────────────

test_that("read_all_snapshot_history returns all datasets without filter", {
  db <- make_test_snapshot_db()
  result <- read_all_snapshot_history(db, n = 100)
  expect_equal(nrow(result), 4L)
  expect_true("ds_a" %in% result$dataset_name)
  expect_true("ds_b" %in% result$dataset_name)
})

test_that("read_all_snapshot_history returns empty frame for missing db", {
  result <- read_all_snapshot_history("/no/such.sqlite", n = 100)
  expect_equal(nrow(result), 0L)
})

# ── Report filename slug construction ─────────────────────────────────────────
# make_report_filename() lives in utils.R and is loaded by setup.R.
# A regression here silently breaks all Open links in server_history.R and ui_datasets.R.

test_that("filename slug produces correct name for standard timestamp", {
  expect_equal(
    make_report_filename("RBB_bonds", "2026-05-31 12:08:08"),
    "RBB_bonds_20260531_120808.html"
  )
})

test_that("filename slug works with T-separator timestamps", {
  expect_equal(
    make_report_filename("my_ds", "2026-05-31T14:28:26"),
    "my_ds_20260531_142826.html"
  )
})

test_that("filename slug works for midnight (all-zero time)", {
  expect_equal(
    make_report_filename("dataset", "2026-01-01 00:00:00"),
    "dataset_20260101_000000.html"
  )
})

test_that("filename slug truncates sub-second precision correctly", {
  # run_dq_check stores timestamps to the second; sub-seconds must be dropped
  expect_equal(
    make_report_filename("ds", "2026-05-31 15:04:34.123456"),
    "ds_20260531_150434.html"
  )
})

test_that("filename slug strips all non-digit characters regardless of separator style", {
  ts_dash  <- "2026-05-31 14:32:21"
  ts_slash <- "2026/05/31 14:32:21"
  ts_T     <- "2026-05-31T14:32:21"
  expect_equal(make_report_filename("ds", ts_dash),  "ds_20260531_143221.html")
  expect_equal(make_report_filename("ds", ts_slash), "ds_20260531_143221.html")
  expect_equal(make_report_filename("ds", ts_T),     "ds_20260531_143221.html")
})

test_that("filename slug handles dataset names with underscores", {
  expect_equal(
    make_report_filename("starwars_folder", "2026-05-31 11:02:03"),
    "starwars_folder_20260531_110203.html"
  )
})

test_that("filename slug vectorises correctly over multiple rows", {
  timestamps <- c("2026-05-31 12:08:08", "2026-05-31 14:28:26")
  ts_raw  <- gsub("[^0-9]", "", substr(timestamps, 1, 19))
  ts_slug <- paste0(substr(ts_raw, 1, 8), "_", substr(ts_raw, 9, 14))
  filenames <- paste0("RBB_bonds_", ts_slug, ".html")
  expect_equal(filenames,
               c("RBB_bonds_20260531_120808.html", "RBB_bonds_20260531_142826.html"))
})

# ── status_badge (htmltools version) ─────────────────────────────────────────

test_that("status_badge returns an htmltools tag for PASS", {
  result <- status_badge("PASS")
  expect_s3_class(result, "shiny.tag")
  html_str <- as.character(result)
  expect_match(html_str, "PASS")
  expect_match(html_str, "#5cb85c")   # green
})

test_that("status_badge returns correct colour for FAIL", {
  html_str <- as.character(status_badge("FAIL"))
  expect_match(html_str, "FAIL")
  expect_match(html_str, "#d9534f")   # red
})

test_that("status_badge returns correct colour for WARN", {
  html_str <- as.character(status_badge("WARN"))
  expect_match(html_str, "WARN")
  expect_match(html_str, "#f0ad4e")   # amber
})

test_that("status_badge returns correct colour for RUNNING", {
  html_str <- as.character(status_badge("RUNNING"))
  expect_match(html_str, "RUNNING")
  expect_match(html_str, "#337ab7")   # blue
})

test_that("status_badge returns a grey badge for unknown status", {
  html_str <- as.character(status_badge("UNKNOWN"))
  expect_match(html_str, "#999999")
})

# Config wizard UI — 8-step framework (spec §8)

wizard_breadcrumb <- function(current, step_valid) {
  labels <- c("Identity","File","Structure","Columns","Rules","Overrides","Custom","Review")
  items <- lapply(seq_along(labels), function(i) {
    cls <- if (i == current) "wizard-step-num active"
           else if (isTRUE(step_valid[i])) "wizard-step-num done"
           else "wizard-step-num disabled"
    tagList(
      span(class=cls, i),
      if (i < length(labels)) span(class="wizard-step-sep", "›")
    )
  })
  div(class="wizard-breadcrumb", items)
}

# ── Step 1: Identity ────────────────────────────────────────────────────
wizard_step1_ui <- function(wiz) {
  tagList(
    h5("Step 1 — Dataset Identity"),
    div(class="mt-3",
      textInput("wiz_dataset_name", "Dataset name *",
                value=wiz$dataset_name, width="400px",
                placeholder="e.g. customer_accounts"),
      p(class="text-muted", style="font-size:12px;",
        "Must start with a letter; only letters, numbers, and underscores.")
    ),
    div(class="mt-2",
      textInput("wiz_description", "Description (optional)",
                value=wiz$description %||% "", width="400px")
    )
  )
}

# ── Step 2: File Location ───────────────────────────────────────────────
wizard_step2_ui <- function(wiz) {
  tagList(
    h5("Step 2 — File Location"),
    div(class="mt-3",
      radioButtons("wiz_file_mode", "How are files identified?",
                   choices=c("Folder scan (automatic)"="folder",
                             "Explicit file paths"="explicit"),
                   selected=wiz$file_mode %||% "folder", inline=TRUE)
    ),
    uiOutput("wiz_file_inputs")
  )
}

# ── Step 3: Format & Structure (CSV or FWF) ─────────────────────────────
wizard_step3_ui <- function() {
  tagList(
    h5("Step 3 — Format and Structure"),
    div(class="mt-3",
      # Character position ruler
      tags$div(id="fwf-char-ruler", class="fwf-char-ruler"),
      # Raw text preview (shinyAce)
      div(id="fwf-ruler-wrap",
        shinyAce::aceEditor("raw_preview", value="",
                            mode="text", readOnly=TRUE,
                            fontSize=13, theme="chrome",
                            height="280px", wordWrap=FALSE)
      ),
      uiOutput("step3_format_hint")
    ),

    div(class="mt-3 p-3 border rounded bg-light",
      h6("File properties"),
      uiOutput("step3_sniff_conflicts"),
      fluidRow(
        column(4,
          radioButtons("wiz_format", "Format",
                       choices=c("CSV"="csv","Fixed-Width (FWF)"="fwf"),
                       selected="csv", inline=TRUE)
        )
      ),
      uiOutput("step3_csv_fields"),
      uiOutput("step3_fwf_fields")
    ),

    div(class="mt-3",
      h6("Parsed preview"),
      uiOutput("step3_no_header_naming"),
      uiOutput("step3_parsed_preview"),
      reactable::reactableOutput("step3_fwf_preview")
    )
  )
}

# ── Step 4: Column Classification ──────────────────────────────────────
wizard_step4_ui <- function() {
  tagList(
    h5("Step 4 — Column Classification"),
    p(class="text-muted", "Review detected columns. Set type overrides, mark key columns, and choose which columns are expected in every delivery."),
    div(class="mb-2 d-flex gap-3 align-items-center",
      tags$span(class="text-muted", style="font-size:12px;", "Expected columns:"),
      actionButton("step4_select_all_expected",  "Select all",  class="btn btn-outline-secondary btn-xs", style="font-size:11px;padding:1px 8px;"),
      actionButton("step4_select_none_expected", "Select none", class="btn btn-outline-secondary btn-xs", style="font-size:11px;padding:1px 8px;")
    ),
    uiOutput("step4_column_table")
  )
}

# ── Step 5: Column Rules ────────────────────────────────────────────────
wizard_step5_ui <- function() {
  tagList(
    h5("Step 5 — Column Rules"),
    p(class="text-muted", "Set optional validation rules per column. Expand a column to add rules. Advanced settings are hidden by default."),
    uiOutput("step5_column_rules")
  )
}

# ── Step 6: Rule Overrides ──────────────────────────────────────────────
wizard_step6_ui <- function(wiz, gcfg) {
  dr <- gcfg$default_rules %||% list()
  overrides <- wiz$rule_overrides %||% list()

  def <- function(key, fallback) overrides[[key]] %||% (dr[[key]] %||% fallback)

  tagList(
    h5("Step 6 — Rule Overrides"),
    p(class="text-muted", "Override global thresholds for this dataset only. Leave unchanged to use global defaults."),
    fluidRow(
      column(4, numericInput("wiz_ro_max_missing",     "Max missing rate",          value=def("max_missing_rate",0.05), min=0,max=1,step=0.01)),
      column(4, numericInput("wiz_ro_max_nonnumeric",  "Max non-numeric rate",      value=def("max_non_numeric_rate",0.01), min=0,max=1,step=0.01)),
      column(4, numericInput("wiz_ro_min_rows",        "Min row count",             value=def("min_row_count",0), min=0,step=1))
    ),
    fluidRow(
      column(4, numericInput("wiz_ro_max_rowchg",      "Max row count change (%)",  value=round(def("max_row_count_change_pct",0.10)*100,2), min=0,step=1)),
      column(4, numericInput("wiz_ro_max_meanshift",   "Max mean shift (%)",        value=round(def("max_numeric_mean_shift_pct",0.20)*100,2), min=0,step=1)),
      column(4, numericInput("wiz_ro_max_misschg",     "Max missing change (pp)",   value=def("max_missing_rate_change_pp",2.0), min=0,step=0.1))
    ),
    fluidRow(
      column(4, numericInput("wiz_ro_max_nonnumchg",   "Max non-numeric change (pp)", value=def("max_non_numeric_rate_change_pp",1.0), min=0,step=0.1)),
      column(4, numericInput("wiz_ro_type_inf",        "Type inference threshold",  value=def("type_inference_threshold",0.90), min=0,max=1,step=0.01))
    ),
    h6("Schema change flags", class="mt-2"),
    fluidRow(
      column(3, checkboxInput("wiz_ro_flag_new",    "Flag new columns",   value=isTRUE(def("flag_new_columns",TRUE)))),
      column(3, checkboxInput("wiz_ro_flag_drop",   "Flag dropped cols",  value=isTRUE(def("flag_dropped_columns",TRUE)))),
      column(3, checkboxInput("wiz_ro_flag_type",   "Flag type changes",  value=isTRUE(def("flag_type_changes",TRUE)))),
      column(3, checkboxInput("wiz_ro_flag_order",  "Flag col order",     value=isTRUE(def("flag_column_order_change",TRUE))))
    )
  )
}

# ── Step 7: Custom Checks ───────────────────────────────────────────────
wizard_step7_ui <- function(wiz) {
  tagList(
    h5("Step 7 — Custom Checks (optional)"),
    p(class="text-muted",
      "Custom checks are organisation-specific rules written in R. They run alongside the standard quality checks."),
    div(class="d-flex gap-2 align-items-center",
      textInput("wiz_custom_file", "Custom checks R file",
                value=wiz$custom_checks_file %||% "", width="400px",
                placeholder="e.g. custom/my_dataset_checks.R"),
      shinyFiles::shinyFilesButton("wiz_custom_browse", "Browse",
        "Select custom checks R file", FALSE,
        style="white-space:nowrap;height:38px;margin-top:24px;")
    ),
    uiOutput("step7_validation_badge"),
    p(class="text-muted mt-2", style="font-size:12px;",
      "The file must define a function named ", tags$code("custom_checks(df)"),
      " or ", tags$code("custom_checks(df, config)"), "."),
    p(class="text-muted", style="font-size:12px;", "Leave blank to skip custom checks.")
  )
}

# ── Step 8: Review and Save ─────────────────────────────────────────────
wizard_step8_ui <- function() {
  tagList(
    h5("Step 8 — Review and Save"),
    uiOutput("step8_summary"),
    div(class="mt-3",
      h6("YAML preview"),
      verbatimTextOutput("yaml_preview"),
    ),
    div(class="mt-3 d-flex gap-2",
      actionButton("wizard_save", "Save config ✓", class="btn btn-success"),
      span(uiOutput("step8_save_status"), style="align-self:center;")
    )
  )
}

# Shared test setup — sourced automatically by testthat before all tests in this dir

library(yaml)
library(withr)

# Source app R files so unit tests can call them directly
.app_r_dir <- system.file("app/R", package = "dqcheckrGUI")
for (.f in list.files(.app_r_dir, pattern = "\\.R$", full.names = TRUE)) {
  source(.f, local = TRUE)
}

# Path to the test fixture CSV (absolute, so configs can reference it)
fixture_csv <- normalizePath(
  file.path(test_path(), "../fixtures/sample.csv"),
  mustWork = TRUE
)

# Build a minimal wiz-like list for build_config_list tests
make_wiz <- function(...) {
  base <- list(
    dataset_name = "test_ds", description = "",
    format = "csv", encoding = "UTF-8", delimiter = ",", quote_char = '"',
    has_header = TRUE, csv_skip = 0L, col_names = character(0),
    raw_header_names = character(0), col_name_reasons = character(0),
    file_mode = "folder", folder = "", current_file = "",
    previous_file = "", fwf_widths = integer(0), fwf_col_names = character(0),
    fwf_skip = 0L, expected_columns = character(0), key_columns = character(0),
    col_types_override = list(), column_rules = list(), rule_overrides = list(),
    custom_checks_file = "", snapshot_db = "", report_output_dir = ""
  )
  args <- list(...)
  for (nm in names(args)) base[[nm]] <- args[[nm]]
  base
}

# Write a YAML fixture to a temp file, return the path
write_yaml_fixture <- function(cfg) {
  path <- tempfile(fileext = ".yml")
  yaml::write_yaml(cfg, path)
  path
}

# Minimal global config suitable for tests (paths don't need to exist for wizard tests)
minimal_global_config <- function(report_dir = tempdir()) {
  list(
    snapshot_db       = file.path(tempdir(), "test_snapshots.sqlite"),
    report_output_dir = report_dir,
    default_rules = list(
      type_inference_threshold       = 0.90,
      max_missing_rate               = 0.05,
      max_non_numeric_rate           = 0.01,
      min_row_count                  = 0L,
      max_row_count_change_pct       = 0.10,
      max_numeric_mean_shift_pct     = 0.20,
      max_missing_rate_change_pp     = 2.0,
      max_non_numeric_rate_change_pp = 1.0,
      flag_new_columns               = TRUE,
      flag_dropped_columns           = TRUE,
      flag_type_changes              = TRUE,
      flag_column_order_change       = TRUE
    )
  )
}

# Create a temporary config directory pre-populated with a global config.
# Returns the path; registers cleanup in the calling test's environment.
make_test_config_dir <- function(envir = parent.frame()) {
  dir <- tempfile(pattern = "dqtest_")
  dir.create(dir, recursive = TRUE)
  withr::defer(unlink(dir, recursive = TRUE), envir = envir)
  yaml::write_yaml(minimal_global_config(), file.path(dir, "dqcheckr.yml"))
  dir
}

# Helper: create AppDriver pointed at the GUI app with a given config dir.
# Stops the driver when the calling test finishes.
make_app_driver <- function(config_dir, envir = parent.frame(), ...) {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  old <- Sys.getenv("DQCHECKR_CONFIG_DIR", unset = NA_character_)
  Sys.setenv(DQCHECKR_CONFIG_DIR = config_dir)
  withr::defer(
    {
      if (is.na(old)) Sys.unsetenv("DQCHECKR_CONFIG_DIR")
      else            Sys.setenv(DQCHECKR_CONFIG_DIR = old)
    },
    envir = envir
  )

  app_dir <- system.file("app", package = "dqcheckrGUI")
  app <- shinytest2::AppDriver$new(app_dir, load_timeout = 30000, ...)
  withr::defer(try(app$stop(), silent = TRUE), envir = envir)
  app
}

# Wait for the wizard Next button to become enabled (step valid).
wait_for_next_enabled <- function(app, timeout_ms = 12000) {
  app$wait_for_js(
    "document.getElementById('wizard_next') !== null &&
     !document.getElementById('wizard_next').hasAttribute('disabled')",
    timeout = timeout_ms
  )
}

# Navigate the wizard from from_step to to_step.
# Waits for Next to be enabled before each click so timing issues don't stall navigation.
open_new_wizard <- function(app) {
  app$click("btn_new_dataset")
  app$wait_for_idle()
}

wizard_go_to_step <- function(app, to_step, from_step = 1L) {
  clicks <- to_step - from_step
  if (clicks <= 0L) return(invisible(NULL))
  for (i in seq_len(clicks)) {
    wait_for_next_enabled(app)
    app$click("wizard_next")
    app$wait_for_idle(timeout = 8000)
  }
  invisible(NULL)
}

# Integration tests: new dataset wizard

library(shinytest2)
library(yaml)

# ── New wizard: folder mode ───────────────────────────────────────────────────

test_that("new wizard (folder mode) saves correct YAML with folder key", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir  <- make_test_config_dir()
  data_dir <- dirname(fixture_csv)
  app      <- make_app_driver(cfg_dir)

  open_new_wizard(app)

  # Step 1
  app$set_inputs(wiz_dataset_name = "folder_test_ds")
  app$wait_for_idle()

  # Step 2: folder mode (default), set path — sleep past the 600ms debounce first
  wizard_go_to_step(app, 2)
  app$set_inputs(wiz_folder_display = data_dir)
  Sys.sleep(0.8)
  app$wait_for_idle(timeout = 6000)

  # Steps 3–8 (already on step 2; navigate the remaining 6 steps)
  wizard_go_to_step(app, 8, from_step = 2)

  # Save
  app$wait_for_js("document.getElementById('wizard_save') !== null", timeout = 5000)
  app$click("wizard_save")
  app$wait_for_idle()

  saved <- file.path(cfg_dir, "folder_test_ds.yml")
  expect_true(file.exists(saved), info = "config file must be created")

  raw <- yaml::read_yaml(saved)
  expect_equal(raw$dataset_name, "folder_test_ds")
  expect_equal(raw$folder,       data_dir)
  expect_null(raw$current_file,
              info = "folder mode must not write current_file")
})

# ── New wizard: explicit file mode ────────────────────────────────────────────
# Regression: file_mode not synced from radio → current_file silently dropped.

test_that("new wizard (explicit mode) saves current_file, not folder", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)

  open_new_wizard(app)

  # Step 1
  app$set_inputs(wiz_dataset_name = "explicit_test_ds")
  app$wait_for_idle()

  # Step 2: explicit mode
  wizard_go_to_step(app, 2)
  app$set_inputs(wiz_file_mode = "explicit")
  app$wait_for_idle()
  app$set_inputs(wiz_current_file_display = fixture_csv)
  Sys.sleep(0.8)
  app$wait_for_idle(timeout = 6000)

  # Steps 3–8 (already on step 2)
  wizard_go_to_step(app, 8, from_step = 2)

  app$wait_for_js("document.getElementById('wizard_save') !== null", timeout = 5000)
  app$click("wizard_save")
  app$wait_for_idle()

  saved <- file.path(cfg_dir, "explicit_test_ds.yml")
  expect_true(file.exists(saved))

  raw <- yaml::read_yaml(saved)
  expect_equal(raw$current_file, fixture_csv,
               info = "current_file must be written in explicit mode")
  expect_null(raw$folder,
              info = "folder must not appear in explicit-mode YAML")
})

# ── Step 1 validation ─────────────────────────────────────────────────────────

test_that("step 1 Next is disabled until a valid name is entered", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)

  open_new_wizard(app)

  # Initially disabled (empty name)
  disabled_initially <- app$get_js(
    "document.getElementById('wizard_next') && document.getElementById('wizard_next').hasAttribute('disabled')"
  )
  expect_true(isTRUE(disabled_initially), info = "Next must be disabled with empty name")

  # Invalid name starting with digit
  app$set_inputs(wiz_dataset_name = "1invalid")
  app$wait_for_idle()
  disabled_invalid <- app$get_js(
    "document.getElementById('wizard_next').hasAttribute('disabled')"
  )
  expect_true(isTRUE(disabled_invalid), info = "Next must stay disabled for invalid name")

  # Valid name — Next must become enabled
  app$set_inputs(wiz_dataset_name = "valid_name")
  app$wait_for_idle()
  disabled_valid <- app$get_js(
    "document.getElementById('wizard_next').hasAttribute('disabled')"
  )
  expect_false(isTRUE(disabled_valid), info = "Next must be enabled for valid name")
})

test_that("step 1 shows duplicate-name error for existing dataset", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(list(dataset_name = "existing_ds", format = "csv"),
                   file.path(cfg_dir, "existing_ds.yml"))

  app <- make_app_driver(cfg_dir)
  open_new_wizard(app)

  app$set_inputs(wiz_dataset_name = "existing_ds")
  app$wait_for_idle()

  page_html <- app$get_html("body")
  expect_true(grepl("already exists", page_html, ignore.case = TRUE))
})

# ── Wizard cancel ─────────────────────────────────────────────────────────────

test_that("cancel wizard discards changes without creating a file", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)

  open_new_wizard(app)
  app$set_inputs(wiz_dataset_name = "cancelled_ds")
  app$wait_for_idle()

  app$click("wizard_cancel")
  app$wait_for_idle()
  app$click("wizard_cancel_confirm")
  app$wait_for_idle()

  expect_false(file.exists(file.path(cfg_dir, "cancelled_ds.yml")),
               info = "cancelled wizard must not write a file")
})

# Integration tests: edit dataset wizard
# Regression suite for the "step N doesn't remember settings" class of bugs.

library(shinytest2)
library(yaml)

make_rich_config <- function(csv_path) {
  list(
    dataset_name     = "rich_test_ds",
    description      = "Integration test dataset",
    format           = "csv",
    encoding         = "UTF-8",
    delimiter        = ",",
    current_file     = csv_path,
    expected_columns = as.list(c("id", "name", "value", "date", "category")),
    key_columns      = as.list(c("id")),
    column_types     = list(value = "numeric"),
    column_rules = list(
      value    = list(min_value = 0, max_value = 1000),
      category = list(allowed_values = as.list(c("A", "B", "C")))
    ),
    rule_overrides   = list(max_missing_rate = 0.10, min_row_count = 5L)
  )
}

open_edit <- function(app, ds) {
  app$run_js(sprintf(
    "Shiny.setInputValue('ds_action',{action:'edit',ds:'%s'},{priority:'event'});", ds
  ))
  app$wait_for_idle(timeout = 5000)
}

nav_to <- function(app, step) {
  wizard_go_to_step(app, step)
}

# ── Step 1 ────────────────────────────────────────────────────────────────────

test_that("edit step 1 restores name and description", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")

  expect_equal(app$get_value(input = "wiz_dataset_name"), "rich_test_ds")
  expect_equal(app$get_value(input = "wiz_description"),  "Integration test dataset")
})

# ── Step 2 ────────────────────────────────────────────────────────────────────

test_that("edit step 2 restores explicit file mode and current_file path", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")
  nav_to(app, 2)

  # Check which radio is checked
  checked_val <- app$get_js(
    "document.querySelector('input[name=\"wiz_file_mode\"]:checked')?.value"
  )
  expect_equal(checked_val, "explicit")

  file_val <- app$get_value(input = "wiz_current_file_display")
  expect_equal(file_val, fixture_csv)
})

test_that("edit step 2 restores folder mode and folder path", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir  <- make_test_config_dir()
  data_dir <- dirname(fixture_csv)
  yaml::write_yaml(
    list(dataset_name = "folder_ds", format = "csv", encoding = "UTF-8",
         delimiter = ",", folder = data_dir),
    file.path(cfg_dir, "folder_ds.yml")
  )
  app <- make_app_driver(cfg_dir)

  open_edit(app, "folder_ds")
  nav_to(app, 2)

  checked_val <- app$get_js(
    "document.querySelector('input[name=\"wiz_file_mode\"]:checked')?.value"
  )
  expect_equal(checked_val, "folder")

  folder_val <- app$get_value(input = "wiz_folder_display")
  expect_equal(folder_val, data_dir)
})

# ── Step 4: column classification ─────────────────────────────────────────────
# Regression: step 4 collector fired before UI rendered, wiping loaded values.

test_that("edit step 4 restores key column checkboxes", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")
  nav_to(app, 4)

  # 'id' is column 1; it should be the key column
  key1_checked <- app$get_js(
    "document.getElementById('s4_key_1')?.checked"
  )
  expect_true(isTRUE(key1_checked), info = "'id' (col 1) must be checked as key column")
})

test_that("edit step 4 restores expected column count", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")
  nav_to(app, 4)

  exp_count <- app$get_js(
    "Array.from(document.querySelectorAll('input[id^=\"s4_exp_\"]')).filter(c=>c.checked).length"
  )
  expect_equal(as.integer(exp_count), 5L, info = "All 5 expected columns must be checked")
})

test_that("edit step 4 restores type overrides", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")
  nav_to(app, 4)

  type_vals <- unlist(app$get_js(
    "Array.from(document.querySelectorAll('select[id^=\"s4_type_\"]')).map(s=>s.value)"
  ))
  expect_true("numeric" %in% type_vals,
              info = "'value' column must have 'numeric' type override restored")
})

# ── Step 5: column rules ──────────────────────────────────────────────────────
# Regression: step 5 collector fired before UI rendered, wiping loaded column_rules.

test_that("edit step 5 restores min/max rules for numeric column", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")
  nav_to(app, 5)

  rule_vals <- unlist(app$get_js(
    "Array.from(document.querySelectorAll('input[id^=\"s5_min_\"],input[id^=\"s5_max_\"]'))
       .filter(i=>i.value!=='').map(i=>parseFloat(i.value))"
  ))
  expect_true(0    %in% rule_vals, info = "min_value=0 must be restored")
  expect_true(1000 %in% rule_vals, info = "max_value=1000 must be restored")
})

# ── Step 6: rule overrides ────────────────────────────────────────────────────
# Regression: collect_step_inputs ran on every Next click, wiping rule_overrides
# loaded from config because step 6 inputs didn't exist yet.

test_that("edit step 6 restores rule overrides from config", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")
  nav_to(app, 6)

  max_missing <- as.numeric(app$get_value(input = "wiz_ro_max_missing"))
  min_rows    <- as.integer(app$get_value(input = "wiz_ro_min_rows"))

  expect_equal(max_missing, 0.10, tolerance = 1e-6,
               info = "max_missing_rate override must be 0.10 from config")
  expect_equal(min_rows, 5L,
               info = "min_row_count override must be 5 from config")
})

# ── Edit → re-save round-trip ────────────────────────────────────────────────

test_that("edit wizard re-saves without corrupting the YAML", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  yaml::write_yaml(make_rich_config(fixture_csv), file.path(cfg_dir, "rich_test_ds.yml"))
  app <- make_app_driver(cfg_dir)

  open_edit(app, "rich_test_ds")
  nav_to(app, 8)

  app$wait_for_js("document.getElementById('wizard_save') !== null", timeout = 5000)
  app$click("wizard_save")
  app$wait_for_idle()

  saved <- yaml::read_yaml(file.path(cfg_dir, "rich_test_ds.yml"))
  expect_equal(saved$dataset_name, "rich_test_ds")
  expect_equal(saved$current_file, fixture_csv)
  expect_equal(unlist(saved$key_columns), "id")
  expect_equal(saved$rule_overrides$max_missing_rate, 0.10, tolerance = 1e-6)
  expect_equal(as.integer(saved$rule_overrides$min_row_count), 5L)
})

# Integration tests: UI behaviour
# Covers interaction bugs found during manual testing.

library(shinytest2)
library(yaml)


# ── Step 1: focus preservation ────────────────────────────────────────────────
# Regression: renderUI re-rendered the whole step on every keystroke, losing focus.

test_that("typing in step 1 does not cause excessive step re-renders", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)
  open_new_wizard(app)

  # Count DOM mutations on the step content container while typing
  app$run_js("
    window.__renderCount = 0;
    const target = document.getElementById('wizard_step_content');
    if (target) new MutationObserver(() => window.__renderCount++)
      .observe(target, { childList: true, subtree: true });
  ")

  for (ch in strsplit("my_dataset", "")[[1]]) {
    app$run_js(sprintf(
      "var el=document.getElementById('wiz_dataset_name');
       el.value+='%s';
       el.dispatchEvent(new Event('input'));", ch
    ))
    Sys.sleep(0.05)
  }
  app$wait_for_idle()

  render_count <- as.integer(app$get_js("window.__renderCount"))
  # Should not re-render 10 times (once per keystroke).
  # Allow a small number for initial binding, but not per-character re-renders.
  expect_lt(render_count, 5L,
            label = paste("step re-render count while typing:", render_count))
})

test_that("step 1 input value survives reactive settle", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)
  open_new_wizard(app)

  app$set_inputs(wiz_dataset_name = "my_test_name")
  app$wait_for_idle()

  expect_equal(app$get_value(input = "wiz_dataset_name"), "my_test_name")
})

# ── Step 4: scroll preservation ───────────────────────────────────────────────
# Regression: renderUI re-built the column table on every checkbox click, scrollTop→0.

test_that("step 4 checkbox click does not reset table scroll position", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir  <- make_test_config_dir()
  data_dir <- dirname(fixture_csv)
  app      <- make_app_driver(cfg_dir)
  open_new_wizard(app)

  app$set_inputs(wiz_dataset_name = "scroll_test")
  app$wait_for_idle()

  wizard_go_to_step(app, 2)
  app$set_inputs(wiz_folder_display = data_dir)
  Sys.sleep(0.8)
  app$wait_for_idle(timeout = 6000)

  wizard_go_to_step(app, 4, from_step = 2)

  # Scroll the column table
  scroll_set <- as.numeric(app$get_js("
    const div = document.querySelector('[style*=\"max-height:500px\"]');
    if (div && div.scrollHeight > div.clientHeight) { div.scrollTop=120; return div.scrollTop; }
    return 0;
  "))

  if (isTRUE(scroll_set > 0)) {
    app$run_js("const cb=document.querySelector('input[id^=\"s4_key_\"]'); if(cb) cb.click();")
    app$wait_for_idle()

    scroll_after <- as.numeric(app$get_js("
      const div=document.querySelector('[style*=\"max-height:500px\"]');
      return div ? div.scrollTop : -1;
    "))
    expect_gt(scroll_after, 0L,
              label = paste("scroll position after checkbox click:", scroll_after))
  } else {
    skip("Table not tall enough to scroll with fixture data — cannot test scroll preservation")
  }
})

# ── Dataset sidebar: sort order ───────────────────────────────────────────────
# Regression: list_dataset_configs returned filesystem order; explicit sort() now added.

test_that("dataset sidebar lists datasets alphabetically regardless of creation order", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  # Create in reverse alphabetical order to stress the sort
  for (nm in c("zebra_ds", "alpha_ds", "mango_ds")) {
    yaml::write_yaml(
      list(dataset_name = nm, format = "csv", folder = dirname(fixture_csv)),
      file.path(cfg_dir, paste0(nm, ".yml"))
    )
    Sys.sleep(0.02)
  }

  app <- make_app_driver(cfg_dir)
  app$wait_for_idle()

  names_in_order <- unlist(app$get_js(
    "Array.from(document.querySelectorAll('.dataset-item span:first-child'))
       .map(s=>s.textContent.trim()).filter(t=>t.length>0)"
  ))

  expect_equal(names_in_order, c("alpha_ds", "mango_ds", "zebra_ds"),
               info = "Sidebar datasets must be alphabetically sorted")
})

# ── Run panel: dataset selector sort order ────────────────────────────────────

test_that("run panel dataset selector is sorted alphabetically", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  for (nm in c("zebra_run", "alpha_run", "mango_run")) {
    yaml::write_yaml(
      list(dataset_name = nm, format = "csv", folder = dirname(fixture_csv)),
      file.path(cfg_dir, paste0(nm, ".yml"))
    )
  }

  app <- make_app_driver(cfg_dir)
  app$run_js("Shiny.setInputValue('nav_run', 1, {priority:'event'});")
  app$wait_for_idle(timeout = 5000)

  # selectize replaces the native <select> with a custom widget; its internal
  # options object is the reliable source of choices in sorted order.
  # The initially-selected value will be datasets[1] from the sorted list.
  selected <- app$get_value(input = "run_dataset")
  expect_equal(selected, "alpha_run",
               info = "Run panel must default to first dataset alphabetically")

  # Also verify all three exist via selectize's internal options object
  selectize_vals <- unlist(app$get_js(
    "Object.keys(document.getElementById('run_dataset').selectize.options)"
  ))
  expect_true(all(c("alpha_run", "mango_run", "zebra_run") %in% selectize_vals),
              info = "All three datasets must appear as selectable options")
})

# ── Step 2: path status badges ────────────────────────────────────────────────

test_that("step 2 shows success badge for valid folder path", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)
  open_new_wizard(app)

  app$set_inputs(wiz_dataset_name = "badge_test")
  app$wait_for_idle()
  wizard_go_to_step(app, 2)

  app$set_inputs(wiz_folder_display = dirname(fixture_csv))
  Sys.sleep(0.8)
  app$wait_for_idle(timeout = 5000)

  badge <- app$get_html("#wiz_folder_path_status")
  expect_true(grepl("Folder found|text-success", badge, ignore.case = TRUE))
})

test_that("step 2 shows error badge for non-existent folder path", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)
  open_new_wizard(app)

  app$set_inputs(wiz_dataset_name = "badge_err")
  app$wait_for_idle()
  wizard_go_to_step(app, 2)

  app$set_inputs(wiz_folder_display = "/path/does/not/exist/xyzabc")
  Sys.sleep(0.8)
  app$wait_for_idle(timeout = 3000)

  badge <- app$get_html("#wiz_folder_path_status")
  expect_true(grepl("not found|text-danger", badge, ignore.case = TRUE))
})

test_that("step 2 explicit mode shows success badge for valid file", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)
  open_new_wizard(app)

  app$set_inputs(wiz_dataset_name = "file_badge_test")
  app$wait_for_idle()
  wizard_go_to_step(app, 2)

  app$set_inputs(wiz_file_mode = "explicit")
  app$wait_for_idle()
  app$set_inputs(wiz_current_file_display = fixture_csv)
  Sys.sleep(0.8)
  app$wait_for_idle(timeout = 3000)

  badge <- app$get_html("#wiz_current_file_status")
  expect_true(grepl("File found|text-success", badge, ignore.case = TRUE))
})

# ── Global config: save and reload ────────────────────────────────────────────

test_that("global config saves values to dqcheckr.yml", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)

  app$run_js("Shiny.setInputValue('nav_global', 1, {priority:'event'});")
  app$wait_for_idle()

  app$set_inputs(
    gcfg_max_missing_rate = 0.42,
    gcfg_snapshot_db      = file.path(cfg_dir, "test.sqlite"),
    gcfg_report_dir       = cfg_dir
  )
  app$wait_for_idle()
  app$click("gcfg_save")
  app$wait_for_idle()

  saved <- yaml::read_yaml(file.path(cfg_dir, "dqcheckr.yml"))
  expect_equal(saved$default_rules$max_missing_rate, 0.42, tolerance = 1e-6)
})

# ── History: Compare button starts disabled ───────────────────────────────────
# Regression: disabled=NA was treated as absent in htmltools, leaving the button
# clickable before any checkboxes were selected.

test_that("history Compare drift button is disabled on initial page load", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir <- make_test_config_dir()
  app     <- make_app_driver(cfg_dir)

  app$run_js("Shiny.setInputValue('nav_history', 1, {priority:'event'});")
  app$wait_for_idle(timeout = 5000)

  disabled <- app$get_js(
    "document.getElementById('history_compare')?.hasAttribute('disabled')"
  )
  expect_true(isTRUE(disabled),
              info = "history_compare button must carry the disabled attribute on load")
})

# ── Dataset panel: Compare button starts disabled ─────────────────────────────

test_that("dataset panel Compare drift button is disabled before checkboxes are checked", {
  skip_if_not_installed("shinytest2")
  skip_if_not_installed("dqcheckr")

  cfg_dir  <- make_test_config_dir()
  data_dir <- dirname(fixture_csv)

  # Create a dataset config so the panel shows last_runs table
  yaml::write_yaml(
    list(dataset_name = "panel_ds", format = "csv", folder = data_dir),
    file.path(cfg_dir, "panel_ds.yml")
  )

  # Write a minimal snapshot row so last_runs is non-empty (table appears)
  db_path <- file.path(cfg_dir, "test.sqlite")
  yaml::write_yaml(
    list(snapshot_db = db_path, report_output_dir = cfg_dir),
    file.path(cfg_dir, "dqcheckr.yml")
  )
  con <- DBI::dbConnect(RSQLite::SQLite(), db_path)
  DBI::dbExecute(con, "
    CREATE TABLE snapshots (
      id INTEGER PRIMARY KEY, dataset_name TEXT, file_name TEXT,
      run_timestamp TEXT, overall_status TEXT,
      check_fail_count INTEGER, check_warn_count INTEGER, row_count INTEGER
    )")
  DBI::dbExecute(con,
    "INSERT INTO snapshots VALUES (1,'panel_ds','f.csv','2026-05-31 10:00:00','PASS',0,0,10)")
  DBI::dbDisconnect(con)

  app <- make_app_driver(cfg_dir)

  # Navigate to the dataset panel
  app$run_js("Shiny.setInputValue('sidebar_dataset_click','panel_ds',{priority:'event'});")
  app$wait_for_idle(timeout = 5000)

  disabled <- app$get_js(
    "document.getElementById('compare_drift_panel_ds')?.hasAttribute('disabled')"
  )
  expect_true(isTRUE(disabled),
              info = "compare_drift button must carry the disabled attribute on load")
})

---
title: "Getting started with dqcheckrGUI"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with dqcheckrGUI}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)
```

`dqcheckrGUI` is a point-and-click Shiny interface for
[dqcheckr](https://github.com/mickmioduszewski/dqcheckr). It lets you
configure dataset quality checks, run them against incoming file deliveries,
and browse historical results — without writing any R code.

The app runs entirely on your local machine. No internet connection, no server,
and no cloud services are required.

## Prerequisites

Install `dqcheckr` before installing `dqcheckrGUI`:

```{r}
install.packages("dqcheckr")
devtools::install_github("mickmioduszewski/dqcheckrGUI")
```

## Launching the app

```{r}
dqcheckrGUI::run_app()
```

The app opens in your default browser at `http://localhost:4321`. It reads and
writes files relative to the working directory you launched from, so set that
to your project folder before calling `run_app()`. The terminal window must
remain open while the app is running — closing it stops the app.

Alternatively, double-click `launch.command` (macOS), `launch.sh` (Linux), or
`launch.bat` (Windows) from the app directory.

### First-run setup

On first launch, if no `dqcheckr.yml` global config file exists in the working
directory, the app opens directly to **Global Config** and prompts you to set
two paths before doing anything else:

- **Snapshot database** — where run summaries are stored (e.g. `data/snapshots.sqlite`)
- **Report output directory** — where HTML reports are written (e.g. `reports/`)

Click **Save global config** once these are set, then proceed to configure
your first dataset.

### Example configuration files

Example YAML configuration files using the Star Wars dataset are included with
the package. Copy them to get started quickly:

```{r}
file.copy(
  system.file("extdata/example_config", package = "dqcheckrGUI"),
  "config",
  recursive = TRUE
)
```

---

## The interface

The app has a fixed sidebar on the left and a main panel that changes based on
your selection.

```
┌──────────────────┬────────────────────────────────┐
│  SIDEBAR         │  MAIN PANEL                    │
│                  │                                │
│  Datasets        │  [content changes here]        │
│    customers ✓   │                                │
│    suppliers ⚠   │                                │
│    [+ New]       │                                │
│                  │                                │
│  ▶ Run           │                                │
│  ⏱ History       │                                │
│  ⚙ Global Config │                                │
└──────────────────┴────────────────────────────────┘
```

### Status badges

Status badges appear in the sidebar, dataset panel, and history table:

| Badge | Meaning |
|-------|---------|
| ✓ PASS | All checks passed |
| ⚠ WARN | One or more warnings; no failures |
| ✗ FAIL | One or more checks failed |
| ● RUNNING | A check is currently in progress |
| — not run | No runs recorded yet |

---

## Dataset panel

Clicking a dataset name in the sidebar opens its **Dataset panel** in the main
area:

```
customer_accounts
──────────────────────────────────────────
Format: CSV   Location: data/incoming/
Config: config/customer_accounts.yml

[Edit config]          [Run check ▶]

Recent runs (last 5):
  2026-05-30   20260530.csv   ✓ PASS   0 failures
  2026-05-23   20260523.csv   ⚠ WARN   0 failures
  2026-05-16   20260516.csv   ✓ PASS   0 failures

[View all in History →]

Compare drift:  ☐ run 1  ☐ run 2   [Compare ▶]
```

- **Edit config** opens the 8-step wizard pre-populated with the existing settings.
- **Run check ▶** navigates to the Run panel with this dataset pre-selected.
- Clicking any row in the recent runs table opens that run's HTML report in a new browser tab.
- **Compare drift** — tick two checkboxes from the recent runs table and click **Compare ▶** to produce a drift report comparing those two deliveries.

---

## Setting up a dataset — the 8-step wizard

Click **+ New dataset** in the sidebar to open the configuration wizard.
Navigate between steps using the **Back** and **Next** buttons, or click the
numbered breadcrumb bar at the top. All values are preserved as you move back
and forth. If you navigate away from the wizard with unsaved changes, the app
will warn you before discarding them.

### Step 1 — Dataset identity

Enter a short machine-readable name for the dataset. Names must start with a
letter and contain only letters, numbers, and underscores (e.g.
`customer_accounts`). This name is used as the config filename and passed
directly to `dqcheckr::run_dq_check()`.

An optional free-text description can also be added.

### Step 2 — File location

Choose how files are identified for each run:

**Folder scan (recommended)** — point to a folder; the app picks the two most
recently modified files automatically on each run. Use this for delivery
processes that drop files into a fixed directory.

A **Preview most recent file** button shows the names and sizes of the two most
recent files in the selected folder: *"Current: 20260530.csv (2.4 MB) | Previous:
20260523.csv (2.3 MB)"*.

**Explicit file paths** — list a current file and optionally a previous file by
path. Useful when files are versioned by name. If no previous file is given,
comparison checks (CP series) are skipped.

### Step 3 — Format and structure

This step identifies the file's format and column layout. The top of the screen
shows a raw text preview of the first 50 lines of the file — the full file is
never loaded.

**Auto-detection**: when a file is loaded, the app uses `readr` to detect the
delimiter, encoding, quote character, and whether the first row is a header. Results are shown as an
editable confirmation panel — you always make the final call, nothing is
silently committed.

#### CSV files

Confirm or adjust:

- **Format** (CSV or Fixed-Width)
- **Delimiter** — comma, tab, semicolon, pipe, space, colon, or custom
- **Encoding** — detected candidates are shown with confidence percentages
- **Quote character**
- **Header row** — whether the first row contains column names

A parsed preview updates live as you change these settings. If there is no
header row, a column-naming panel appears alongside the preview where you can
enter names for each column. Names must be valid R identifiers; a suggestion
is offered if you enter something invalid.

#### Fixed-width (FWF) files

When FWF is selected, a **visual ruler** activates above the text preview.

- **Click** anywhere in the text to place a column boundary at that character
  position.
- **Drag** a boundary line left or right to adjust it — it snaps to character
  boundaries.
- **Double-click** a boundary line to remove it.
- An **Auto-detect** button attempts automatic boundary detection using
  `readr::fwf_empty()`.

Below the ruler, a table shows the resulting column definitions (start
position, width, name, type). Edit column names and types here. A validation
badge confirms whether the column widths account for the full record length:

- ✅ green — widths match the line length exactly
- ⚠ amber — widths cover fewer characters than the line length (trailing
  columns unaccounted for)
- ✗ red — widths exceed the line length (Next is disabled until fixed)

If the file has header rows to skip before the data begins, set
**Header rows to skip**. Column names are pre-populated from the skipped header
row if one is present.

### Step 4 — Column classification

Review every detected column. For each column you can:

- **Set a type override** — force a column to be treated as `character` even
  if it looks numeric. Use this for postcodes, phone numbers, account codes,
  BSB numbers, or any identifier that happens to contain only digits. Click the
  inferred type to open the override dropdown.
- **Mark as key column** — that column will be checked for uniqueness (QC-12).
- **Mark as expected** — columns not present in a delivery will raise a schema
  warning (SC-01). All columns are expected by default; use **Select none** to
  clear all and tick only the columns that must always be present.

### Step 5 — Column rules

Each column is shown as a collapsible card. Expand a column to add optional
per-column validation rules.

**Standard rules** (always visible):

| Rule | Applies to | Check |
|------|------------|-------|
| Allowed values | character columns | QC-09: flag any value not in the list |
| Min value | numeric columns | QC-10: flag values below the minimum |
| Max value | numeric columns | QC-10: flag values above the maximum |

**Advanced rules** (click **Advanced ▼** to reveal):

| Rule | Description |
|------|-------------|
| Regex pattern | Flag values that do not match the pattern (QC-13). Click **Test against sample** to verify the pattern against the actual file before saving. |
| Max missing rate | Override the dataset-level threshold for this column only |
| Max non-numeric rate | Override for numeric columns only |
| Max missing rate change | Override the comparison threshold for this column |
| Max mean shift | Override the mean shift threshold for this column |

A regex syntax error disables **Next**; a pattern that fails against sample
values shows a warning but does not block you from proceeding.

### Step 6 — Rule overrides

Override the global default thresholds for this dataset only. Each field is
pre-filled with the current global default. Only fields whose value differs from
the global default are written to the dataset config; fields left at their
default are omitted so that a later change to the global config is automatically
inherited.

| Threshold | Default | What it controls |
|-----------|---------|-----------------|
| Max missing rate | 0.05 | Flag a column if > 5 % of values are blank |
| Max non-numeric rate | 0.01 | Flag a numeric column if > 1 % of values cannot be parsed |
| Min row count | 0 (off) | Fail if the delivery has fewer rows than this |
| Max row count change | 10 % | Warn if row count changes by > 10 % vs previous delivery |
| Max mean shift | 20 % | Warn if a numeric column mean shifts by > 20 % |
| Max missing change | 2 pp | Warn if missing rate changes by > 2 percentage points |
| Max non-numeric change | 1 pp | Warn if non-numeric rate changes by > 1 percentage point |
| Type inference threshold | 0.90 | A column is typed numeric if ≥ 90 % of values parse as numbers |

Schema change flags control whether warnings are raised when columns are added,
dropped, change type, or change order between deliveries.

### Step 7 — Custom checks (optional)

Point to a plain R file that defines a `custom_checks(df)` function. The app
validates the file immediately: it checks that the file exists, parses without
syntax errors, and defines the expected function. A green badge confirms a
valid file; a red badge shows the specific problem.

Leave this field blank to skip custom checks. See
`vignette("dqcheckr", package = "dqcheckr")` for the custom checks function
signature and return value.

### Step 8 — Review and save

A summary of all settings is shown alongside a YAML preview of the
configuration that will be written. Click **Save config ✓** to write the file.
A success notification confirms the path. The app then navigates to the dataset
panel for the saved dataset.

**For analysts who hand-edit YAML**: the app preserves any keys you have added
to the YAML file outside the wizard. On the next edit, those keys appear in
the Step 8 preview under `# preserved from original file` and are written back
unchanged. The wizard never silently drops hand-added config keys.

---

## Running a quality check

Click **▶ Run** in the sidebar (or **Run check ▶** from a dataset panel).

1. Select a dataset from the dropdown.
2. The app validates the config before enabling the run button — it checks that
   the config file is readable, the data folder or files exist, and the output
   paths are writable. Any problem is shown as a red badge with the specific
   issue.
3. Click **▶ Run check**.

The check runs in a background process so the UI stays responsive. Progress is
streamed to the log area in real time. When complete:

```
Status: ✓ PASS    [Open report ↗]   [View log]

  0 failures   0 warnings   22 passed
  Report: reports/customer_accounts_20260531_143022.html
```

- **Open report ↗** opens the HTML report in the system browser.
- **View log** shows the full run log in a dialog box.
- **■ Stop** (shown during a run) cancels the background process. A
  confirmation is shown first, noting that the snapshot will not be written if
  the run has not completed.

---

## Browsing history

Click **⏱ History** in the sidebar to see all past runs across all datasets.

The table can be filtered by dataset name, status, or date using the filter row
at the top of each column. Click **Load more** to page through older runs.

**To open a report**: click the **Open** link in the Report column to open that
run's HTML report in a new browser tab.

**To compare two deliveries**:

1. Tick the checkbox next to exactly two runs **from the same dataset**.
2. Click **Compare drift ▶**.

The drift report opens in a new tab, showing column-by-column changes between
the two snapshots. The button is disabled with a tooltip if the selected rows
are from different datasets.

---

## Global configuration

Click **⚙ Global Config** to set defaults that apply to all datasets unless
overridden at the dataset level (Step 6).

**Infrastructure paths** — set the snapshot database path and report output
directory. The snapshot database is created automatically on first run if it
does not exist. If either path's parent directory does not exist, a red
validation message is shown — create the directory on disk first, then save.

**Default rule thresholds** — the same thresholds described in Step 6 above,
applied globally. Dataset-level overrides take precedence.

Click **Save global config** to write changes.

### Team use via a shared network drive

Multiple users can share the same configuration and run history by pointing
their `config_dir` and snapshot database paths at a shared network folder. Each
user runs the app locally; the shared drive holds the YAML configs and the
SQLite snapshot database. Set these paths in Global Config on each user's
machine. SQLite handles concurrent reads safely; two simultaneous writes
(unlikely in practice) are serialised by file locking.

---

## Learn more

See `vignette("dqcheckr", package = "dqcheckr")` for a full description of
every quality check (QC-01 to QC-14, SC-01/02, CP-01 to CP-08), per-column
configuration options, custom checks, snapshot database schema, and the
`compare_snapshots()` function used for drift reports.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_app.R
\name{run_app}
\alias{run_app}
\title{Launch the dqcheckrGUI Shiny application}
\usage{
run_app(...)
}
\arguments{
\item{...}{Arguments passed to \code{\link[shiny]{runApp}} (e.g.
\code{port}, \code{launch.browser}).}
}
\value{
Called for its side effect; does not return a value.
}
\description{
Opens the point-and-click interface for configuring and running
\pkg{dqcheckr} dataset quality checks.
}
\examples{
if (interactive()) {
  run_app()
}
}
